import botocore.client
import botocore.exceptions
import contextlib
import datetime
from .history_handler import history
from awslabs.aws_api_mcp_server.core.common.models import Credentials
from copy import deepcopy
from unittest.mock import MagicMock, patch


S3_CLI_NO_REGION = 'aws s3api list-buckets'

CLOUD9_PARAMS_CLI_MISSING_CONTEXT = (
    'aws cloud9 create-environment-ec2 --name test --instance-type t3.large'
)
CLOUD9_PARAMS_MISSING_CONTEXT_FAILURES = {
    'validation_failures': None,
    'missing_context_failures': [
        {
            'reason': "The following parameters are missing for service 'cloud9' and operation 'create-environment-ec2': '--image-id'",
            'context': {
                'service': 'cloud9',
                'operation': 'create-environment-ec2',
                'parameters': ['--image-id'],
                'args': None,
                'region': None,
                'operators': None,
            },
        }
    ],
}

CLOUD9_PARAMS_CLI_NON_EXISTING_OPERATION = 'aws cloud9 list-environments-1'
CLOUD9_PARAMS_CLI_VALIDATION_FAILURES = {
    'validation_failures': [
        {
            'reason': "The operation 'list-environments-1' for service 'cloud9' does not exist.",
            'context': {
                'service': 'cloud9',
                'operation': 'list-environments-1',
                'parameters': None,
                'args': None,
                'region': None,
                'operators': None,
            },
        }
    ],
    'missing_context_failures': None,
}


TEST_CREDENTIALS = {'access_key_id': 'test', 'secret_access_key': 'test', 'session_token': 'test'}


# Original botocore _make_api_call function
orig = botocore.client.BaseClient._make_api_call

CLOUD9_LIST_ENVIRONMENTS = {
    'environmentIds': [
        'dc7ec5068da34567b72376837becd583',  # pragma: allowlist secret
        'bfdc3c72123b4b918de2004b6d6e78ab',  # pragma: allowlist secret
    ],  # pragma: allowlist secret
    'ResponseMetadata': {'HTTPStatusCode': 200},
}

T3_EC2_INSTANCE_RESERVATION = {
    'Groups': [],
    'Instances': [
        {
            'AmiLaunchIndex': 0,
            'ImageId': 'ami-0a5fbecff26409d48',
            'InstanceId': 'i-0487c758efa644ff4',
            'InstanceType': 't3.small',
            'LaunchTime': datetime.datetime.now(datetime.timezone.utc),
            'Monitoring': {'State': 'disabled'},
        }
    ],
}

T2_EC2_INSTANCE_RESERVATION = {
    'Groups': [],
    'Instances': [
        {
            'AmiLaunchIndex': 1,
            'ImageId': 'ami-0a5fbecff26409d48',
            'InstanceId': 'i-0487c758efa644ff4',
            'InstanceType': 't2.micro',
            'LaunchTime': datetime.datetime.now(datetime.timezone.utc),
            'Monitoring': {'State': 'enabled'},
        }
    ],
}

EMPTY_EC2_RESERVATION = {'Groups': [], 'Instances': []}

T2_EC2_DESCRIBE_INSTANCES_FILTERED = {
    'Result': [
        EMPTY_EC2_RESERVATION.get('Instances'),
        T2_EC2_INSTANCE_RESERVATION.get('Instances'),
    ],
    'ResponseMetadata': {'HTTPStatusCode': 200},
}

EC2_DESCRIBE_INSTANCES = {
    'Reservations': [
        T3_EC2_INSTANCE_RESERVATION,
        T2_EC2_INSTANCE_RESERVATION,
    ],
    'ResponseMetadata': {'HTTPStatusCode': 200},
}

CLOUD9_DESCRIBE_ENVIRONMENTS = {
    'environments': [
        {'id': '7d61007bd98b4d589f1504af84c168de'},  # pragma: allowlist secret
        {'id': 'b181ffd35fe2457c8c5ae9d75edc068a'},  # pragma: allowlist secret
    ],
    'ResponseMetadata': {'HTTPStatusCode': 200},
}

GET_CALLER_IDENTITY_PAYLOAD = {
    'ResponseMetadata': {'HTTPStatusCode': 200},
}

IAD_BUCKET = {'Name': 'IAD', 'CreationDate': '2022-07-13T15:20:58+00:00'}
DUB_BUCKET = {'Name': 'DUB', 'CreationDate': '2022-07-13T15:20:58+00:00'}
PDX_BUCKET = {'Name': 'PDX', 'CreationDate': '2022-07-13T15:20:58+00:00'}

LIST_BUCKETS_PAYLOAD = {
    'ResponseMetadata': {'HTTPStatusCode': 200},
    'Buckets': [
        IAD_BUCKET,
        DUB_BUCKET,
        PDX_BUCKET,
    ],
    'Owner': {'DisplayName': 'clpo', 'ID': '***'},
}

SSM_LIST_NODES_PAYLOAD = {
    'Nodes': [
        {
            'CaptureTime': '1970-01-01T00:00:00',
            'Id': 'abc',
            'NodeType': {
                'Instance': {
                    'AgentType': 'AgentType',
                    'AgentVersion': 'AgentVersion',
                    'ComputerName': 'ComputerName',
                    'InstanceStatus': 'InstanceStatus',
                    'IpAddress': 'IpAddress',
                    'ManagedStatus': 'ManagedStatus',
                    'PlatformType': 'PlatformType',
                    'PlatformName': 'PlatformName',
                    'PlatformVersion': 'PlatformVersion',
                    'ResourceType': 'ResourceType',
                }
            },
        }
    ],
    'ResponseMetadata': {'HTTPStatusCode': 200},
    'estimated_resources_processed': 20,
}

CLOUDFRONT_FUNCTIONS = {
    'ResponseMetadata': {'HTTPStatusCode': 200},
    'FunctionList': {
        'Items': [
            {
                'Name': 'my-function-1',
            },
            {
                'Name': 'my-function-2',
            },
        ],
        'MaxItems': 10,
    },
}


class MockStreamingBody:
    """Mock implementation of boto3's StreamingBody for testing."""

    def __init__(self, content):
        """Initialize the mock streaming body with content."""
        self.content = content

    def iter_chunks(self, chunk_size=1024):
        """Yield chunks of the content."""
        yield self.content

    def read(self):
        """Read all content."""
        return self.content


S3_GET_OBJECT_PAYLOAD = {
    'Body': MockStreamingBody(b'test content'),
    'ResponseMetadata': {'HTTPStatusCode': 200},
}

LAMBDA_INVOKE_PAYLOAD = {
    'Payload': MockStreamingBody(b'{"result": "success"}'),
    'ResponseMetadata': {'HTTPStatusCode': 200},
}


def raise_(ex):
    """Raise the given exception."""
    raise ex


_patched_operations = {
    'ListEnvironments': lambda *args, **kwargs: CLOUD9_LIST_ENVIRONMENTS,
    'DescribeInstances': lambda *args, **kwargs: deepcopy(EC2_DESCRIBE_INSTANCES),
    'ListFunctions': lambda *args, **kwargs: CLOUDFRONT_FUNCTIONS,
    'DescribeEnvironments': lambda *args, **kwargs: CLOUD9_DESCRIBE_ENVIRONMENTS,
    'GetCallerIdentity': lambda *args, **kwargs: GET_CALLER_IDENTITY_PAYLOAD,
    'ListBuckets': lambda *args, **kwargs: LIST_BUCKETS_PAYLOAD,
    'ListNodes': lambda *args, **kwargs: SSM_LIST_NODES_PAYLOAD,
    'GetObject': lambda *args, **kwargs: S3_GET_OBJECT_PAYLOAD,
    'Invoke': lambda *args, **kwargs: LAMBDA_INVOKE_PAYLOAD,
}


def mock_make_api_call(self, operation_name, kwarg):
    """Mock the _make_api_call method for boto3 clients."""
    op = _patched_operations.get(operation_name)
    if op:
        history.emit(
            operation_name,
            kwarg,
            self._client_config.region_name,
            self._client_config.read_timeout,
            self.meta.endpoint_url,
        )
        return op(kwarg)

    # If we don't want to patch the API call; these will fail
    # as credentials are invalid
    return orig(self, operation_name, kwarg)


def create_file_open_mock(*target_files):
    """Create a mock open function that only mocks specific files."""
    original_open = open
    mock_files = {filename: MagicMock() for filename in target_files}

    def mock_open_side_effect(filename, mode='r', *args, **kwargs):
        if filename in target_files:
            mock_context = MagicMock()
            mock_file = mock_files[filename]
            mock_context.__enter__.return_value = mock_file
            return mock_context
        return original_open(filename, mode, *args, **kwargs)

    return mock_open_side_effect, mock_files


@contextlib.contextmanager
def patch_boto3():
    """Context manager to patch boto3 for non-paginated API calls."""
    with patch(
        'awslabs.aws_api_mcp_server.core.aws.driver.get_local_credentials',
        return_value=Credentials(**TEST_CREDENTIALS),
    ):
        with patch_botocore():
            yield


@contextlib.contextmanager
def patch_botocore():
    """Patch botocore."""

    def mock_can_paginate(self, operation_name):
        return False

    with patch('botocore.client.BaseClient._make_api_call', new=mock_make_api_call):
        with patch('botocore.client.BaseClient.can_paginate', new=mock_can_paginate):
            yield


class DummyCtx:
    """Mock implementation of MCP context for testing purposes."""

    async def error(self, message):
        """Mock MCP ctx.error with the given message.

        Args:
            message: The error message
        """
        # Do nothing because MCP ctx.error doesn't throw exception
        pass



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .services import PaginationConfig
from botocore.paginate import PageIterator, Paginator
from botocore.utils import merge_dicts, set_value_from_jmespath
from jmespath.parser import ParsedResult
from loguru import logger
from typing import Any


def _merge_page_into_result(
    result: dict[str, Any],
    page: dict[str, Any],
    page_iterator: PageIterator,
) -> dict[str, Any]:
    for result_expression in page_iterator.result_keys:
        result_value = result_expression.search(page)
        if result_value is None:
            continue

        existing_value = result_expression.search(result)
        if existing_value is None:
            # Set the initial result
            set_value_from_jmespath(
                result,
                result_expression.expression,
                result_value,
            )
            continue

        # Merge with existing value
        if isinstance(result_value, list):
            existing_value.extend(result_value)
        elif isinstance(result_value, (int | float | str)):
            # Modify the existing result with the sum or concatenation
            set_value_from_jmespath(
                result,
                result_expression.expression,
                existing_value + result_value,
            )

    return result


def _finalize_result(
    result: dict[str, Any],
    page_iterator: PageIterator,
    response_metadata: dict[str, Any] | None,
    client_side_filter: ParsedResult | None,
) -> dict[str, Any]:
    """Finalize the result by adding non-aggregate parts and processing metadata."""
    if client_side_filter is not None:
        # Apply client-side filter
        result = {'Result': client_side_filter.search(result)}

    merge_dicts(result, page_iterator.non_aggregate_part)

    result['ResponseMetadata'] = response_metadata

    if page_iterator.resume_token is not None:
        result['pagination_token'] = page_iterator.resume_token

    return result


def build_result(
    paginator: Paginator,
    service_name: str,
    operation_name: str,
    operation_parameters: dict[str, Any],
    pagination_config: PaginationConfig,
    client_side_filter: ParsedResult | None = None,
):
    """This function is based on build_full_result in botocore with some modifications.

    to take into account token limits, max results and timeouts. The first page is always processed.

    https://github.com/boto/botocore/blob/master/botocore/paginate.py#L481
    """
    result: dict[str, Any] = {}
    response_metadata = None

    logger.info(
        f'Building pagination result for {service_name} {operation_name} with config: {pagination_config}'
    )
    page_iterator = paginator.paginate(**operation_parameters, PaginationConfig=pagination_config)

    for response in page_iterator:
        page = response

        # operation object pagination comes in a tuple of two elements: (http_response, parsed_response)
        if isinstance(response, tuple) and len(response) == 2:
            page = response[1]

        # For each page in the response we need to inject the necessary components from the page into the result.
        _merge_page_into_result(result, page, page_iterator)

        response_metadata = page.get('ResponseMetadata')

    return _finalize_result(result, page_iterator, response_metadata, client_side_filter)



import pytest
import requests
from awslabs.aws_api_mcp_server.core.common.errors import AwsApiMcpError, CommandValidationError
from awslabs.aws_api_mcp_server.core.common.help_command import generate_help_document
from awslabs.aws_api_mcp_server.core.common.helpers import as_json
from awslabs.aws_api_mcp_server.core.common.models import (
    AwsCliAliasResponse,
    Consent,
    Credentials,
    InterpretationResponse,
    ProgramInterpretationResponse,
)
from awslabs.aws_api_mcp_server.server import call_aws, call_aws_helper, main, suggest_aws_commands
from botocore.exceptions import NoCredentialsError
from fastmcp.server.elicitation import AcceptedElicitation
from tests.fixtures import TEST_CREDENTIALS, DummyCtx
from unittest.mock import AsyncMock, MagicMock, patch


@patch('awslabs.aws_api_mcp_server.server.get_read_only_operations')
@patch('awslabs.aws_api_mcp_server.server.server')
def test_main_read_operations_index_load_failure(mock_server, mock_get_read_ops):
    """Test main function when read operations index loading fails."""
    mock_get_read_ops.side_effect = Exception('Failed to load operations')

    with patch('awslabs.aws_api_mcp_server.server.WORKING_DIRECTORY', '/tmp/test'):
        with patch('awslabs.aws_api_mcp_server.server.DEFAULT_REGION', 'us-east-1'):
            with patch('awslabs.aws_api_mcp_server.server.validate_aws_region'):
                # Should not raise exception, just log warning
                main()
                mock_server.run.assert_called_once()


@patch('awslabs.aws_api_mcp_server.server.DEFAULT_REGION', 'us-east-1')
@patch('awslabs.aws_api_mcp_server.server.interpret_command')
@patch('awslabs.aws_api_mcp_server.server.validate')
@patch('awslabs.aws_api_mcp_server.server.translate_cli_to_ir')
@patch('awslabs.aws_api_mcp_server.core.aws.service.is_operation_read_only')
async def test_call_aws_success(
    mock_is_operation_read_only,
    mock_translate_cli_to_ir,
    mock_validate,
    mock_interpret,
):
    """Test call_aws returns success for a valid read-only command."""
    # Create a proper ProgramInterpretationResponse mock
    mock_response = InterpretationResponse(error=None, json='{"Buckets": []}', status_code=200)

    mock_result = ProgramInterpretationResponse(
        response=mock_response,
        metadata=None,
        validation_failures=None,
        missing_context_failures=None,
        failed_constraints=None,
    )
    mock_interpret.return_value = mock_result

    mock_is_operation_read_only.return_value = True

    # Mock IR with command metadata
    mock_ir = MagicMock()
    mock_ir.command_metadata = MagicMock()
    mock_ir.command_metadata.service_sdk_name = 's3api'
    mock_ir.command_metadata.operation_sdk_name = 'list-buckets'
    mock_ir.command = MagicMock()
    mock_ir.command.is_awscli_customization = False  # Ensure interpret_command is called
    mock_ir.command.is_help_operation = False
    mock_ir.command.is_help_operation = False
    mock_translate_cli_to_ir.return_value = mock_ir

    mock_response = MagicMock()
    mock_response.validation_failed = False
    mock_validate.return_value = mock_response

    # Execute
    result = await call_aws.fn('aws s3api list-buckets', DummyCtx())

    # Verify - the result should be the ProgramInterpretationResponse object
    assert result == mock_result
    mock_translate_cli_to_ir.assert_called_once_with('aws s3api list-buckets')
    mock_validate.assert_called_once_with(mock_ir)
    mock_interpret.assert_called_once()


@patch('awslabs.aws_api_mcp_server.server.get_requests_session')
async def test_suggest_aws_commands_success(mock_get_session):
    """Test suggest_aws_commands returns suggestions for a valid query."""
    mock_suggestions = {
        'suggestions': [
            {
                'command': 'aws s3 ls',
                'confidence': 0.95,
                'description': 'List S3 buckets',
                'required_parameters': [],
            },
            {
                'command': 'aws s3api list-buckets',
                'confidence': 0.90,
                'description': 'List all S3 buckets using S3 API',
                'required_parameters': [],
            },
        ]
    }

    mock_response = MagicMock()
    mock_response.json.return_value = mock_suggestions

    mock_session = MagicMock()
    mock_session.post.return_value = mock_response
    mock_session.__enter__.return_value = mock_session
    mock_session.__exit__.return_value = None

    mock_get_session.return_value = mock_session

    result = await suggest_aws_commands.fn('List all S3 buckets', DummyCtx())

    assert result == mock_suggestions
    mock_session.post.assert_called_once()

    # Verify the HTTP call parameters
    call_args = mock_session.post.call_args
    assert call_args[1]['json'] == {'query': 'List all S3 buckets'}
    assert call_args[1]['timeout'] == 30


async def test_suggest_aws_commands_empty_query():
    """Test suggest_aws_commands raises error for empty query."""
    with pytest.raises(AwsApiMcpError) as exc_info:
        await suggest_aws_commands.fn('', DummyCtx())

    assert 'Empty query provided' in str(exc_info.value)


@patch('awslabs.aws_api_mcp_server.server.get_requests_session')
async def test_suggest_aws_commands_exception(mock_get_session):
    """Test suggest_aws_commands raises error when HTTPError is raised."""
    mock_response = MagicMock()
    mock_response.raise_for_status.side_effect = requests.HTTPError('404 Not Found')

    mock_session = MagicMock()
    mock_session.post.return_value = mock_response
    mock_session.__enter__.return_value = mock_session
    mock_session.__exit__.return_value = None

    mock_get_session.return_value = mock_session

    with pytest.raises(AwsApiMcpError) as exc_info:
        await suggest_aws_commands.fn('List S3 buckets', DummyCtx())

    assert 'Failed to execute tool due to internal error' in str(exc_info.value)
    mock_response.raise_for_status.assert_called_once()
    mock_session.post.assert_called_once()


@patch('awslabs.aws_api_mcp_server.server.execute_awscli_customization')
@patch('awslabs.aws_api_mcp_server.server.validate')
@patch('awslabs.aws_api_mcp_server.server.translate_cli_to_ir')
async def test_call_aws_helper_passes_region_to_customization(
    mock_translate_cli_to_ir,
    mock_validate,
    mock_execute,
):
    """Ensure region is forwarded to execute_awscli_customization when using customizations."""
    # Arrange IR with customization flag
    mock_command = MagicMock()
    mock_command.is_awscli_customization = True
    mock_command.is_help_operation = False
    mock_command.service_name = 's3'
    mock_command.operation_cli_name = 'ls'
    mock_ir = MagicMock()
    mock_ir.command = mock_command
    mock_translate_cli_to_ir.return_value = mock_ir

    mock_validate.return_value = MagicMock(validation_failed=False)
    mock_execute.return_value = AwsCliAliasResponse(response='', error='')

    # Avoid policy gating
    with (
        patch('awslabs.aws_api_mcp_server.server.READ_OPERATIONS_INDEX', None),
        patch('awslabs.aws_api_mcp_server.server.READ_OPERATIONS_ONLY_MODE', False),
        patch('awslabs.aws_api_mcp_server.server.REQUIRE_MUTATION_CONSENT', False),
    ):
        # Act
        result = await call_aws_helper(
            cli_command='aws s3 ls',
            ctx=DummyCtx(),  # type: ignore[arg-type]
            max_results=None,
            credentials=None,
            default_region='eu-west-1',
        )

    # Assert
    assert isinstance(result, AwsCliAliasResponse)
    _, kwargs = mock_execute.call_args
    assert kwargs.get('default_region_override') == 'eu-west-1'


@patch('awslabs.aws_api_mcp_server.server.interpret_command')
@patch('awslabs.aws_api_mcp_server.server.validate')
@patch('awslabs.aws_api_mcp_server.server.translate_cli_to_ir')
async def test_call_aws_helper_passes_region_to_interpret(
    mock_translate_cli_to_ir,
    mock_validate,
    mock_interpret,
):
    """Ensure region is forwarded to interpret_command for normal operations."""
    # Arrange IR without customization
    mock_command = MagicMock()
    mock_command.is_awscli_customization = False
    mock_command.is_help_operation = False
    mock_command.service_name = 's3api'
    mock_command.operation_cli_name = 'list-buckets'
    mock_ir = MagicMock()
    mock_ir.command = mock_command
    mock_translate_cli_to_ir.return_value = mock_ir

    mock_validate.return_value = MagicMock(validation_failed=False)
    mock_interpret.return_value = ProgramInterpretationResponse(
        response=InterpretationResponse(error=None, json='{}', status_code=200),
        metadata=None,
        validation_failures=None,
        missing_context_failures=None,
        failed_constraints=None,
    )

    # Avoid policy gating
    with (
        patch('awslabs.aws_api_mcp_server.server.READ_OPERATIONS_INDEX', None),
        patch('awslabs.aws_api_mcp_server.server.READ_OPERATIONS_ONLY_MODE', False),
        patch('awslabs.aws_api_mcp_server.server.REQUIRE_MUTATION_CONSENT', False),
    ):
        # Act
        result = await call_aws_helper(
            cli_command='aws s3api list-buckets',
            ctx=DummyCtx(),  # type: ignore[arg-type]
            max_results=None,
            credentials=None,
            default_region='eu-west-2',
        )

    # Assert
    assert isinstance(result, ProgramInterpretationResponse)
    _, kwargs = mock_interpret.call_args
    assert kwargs.get('default_region_override') == 'eu-west-2'


@patch('awslabs.aws_api_mcp_server.server.DEFAULT_REGION', 'us-east-1')
@patch('awslabs.aws_api_mcp_server.server.REQUIRE_MUTATION_CONSENT', True)
@patch('awslabs.aws_api_mcp_server.server.interpret_command')
@patch('awslabs.aws_api_mcp_server.server.validate')
@patch('awslabs.aws_api_mcp_server.server.translate_cli_to_ir')
@patch('awslabs.aws_api_mcp_server.core.aws.service.is_operation_read_only')
async def test_call_aws_with_consent_and_accept(
    mock_is_operation_read_only,
    mock_translate_cli_to_ir,
    mock_validate,
    mock_interpret,
):
    """Test call_aws with mutating action and consent enabled."""
    # Create a proper ProgramInterpretationResponse mock
    mock_response = InterpretationResponse(error=None, json='{"Buckets": []}', status_code=200)

    mock_result = ProgramInterpretationResponse(
        response=mock_response,
        metadata=None,
        validation_failures=None,
        missing_context_failures=None,
        failed_constraints=None,
    )
    mock_interpret.return_value = mock_result

    mock_is_operation_read_only.return_value = False

    # Mock IR with command metadata
    mock_ir = MagicMock()
    mock_ir.command_metadata = MagicMock()
    mock_ir.command_metadata.service_sdk_name = 's3api'
    mock_ir.command_metadata.operation_sdk_name = 'create-bucket'
    mock_ir.command = MagicMock()
    mock_ir.command.is_awscli_customization = False  # Ensure interpret_command is called
    mock_ir.command.is_help_operation = False
    mock_ir.command.is_help_operation = False
    mock_translate_cli_to_ir.return_value = mock_ir

    mock_response = MagicMock()
    mock_response.validation_failed = False
    mock_validate.return_value = mock_response

    mock_ctx = AsyncMock()
    mock_ctx.elicit.return_value = AcceptedElicitation(data=Consent(answer=True))

    # Execute
    result = await call_aws.fn('aws s3api create-bucket --bucket somebucket', mock_ctx)

    # Verify that consent was requested
    assert result == mock_result
    mock_translate_cli_to_ir.assert_called_once_with('aws s3api create-bucket --bucket somebucket')
    mock_validate.assert_called_once_with(mock_ir)
    mock_interpret.assert_called_once()
    mock_ctx.elicit.assert_called_once()


@patch('awslabs.aws_api_mcp_server.server.DEFAULT_REGION', 'us-east-1')
@patch('awslabs.aws_api_mcp_server.server.REQUIRE_MUTATION_CONSENT', True)
@patch('awslabs.aws_api_mcp_server.server.interpret_command')
@patch('awslabs.aws_api_mcp_server.server.validate')
@patch('awslabs.aws_api_mcp_server.server.translate_cli_to_ir')
@patch('awslabs.aws_api_mcp_server.core.aws.service.is_operation_read_only')
async def test_call_aws_with_consent_and_reject(
    mock_is_operation_read_only,
    mock_translate_cli_to_ir,
    mock_validate,
    mock_interpret,
):
    """Test call_aws with mutating action and consent enabled."""
    mock_response = InterpretationResponse(error=None, json='{"Buckets": []}', status_code=200)
    mock_is_operation_read_only.return_value = False

    # Mock IR with command metadata
    mock_ir = MagicMock()
    mock_ir.command_metadata = MagicMock()
    mock_ir.command_metadata.service_sdk_name = 's3api'
    mock_ir.command_metadata.operation_sdk_name = 'create-bucket'
    mock_ir.command = MagicMock()
    mock_ir.command.is_awscli_customization = False
    mock_ir.command.is_help_operation = False
    mock_translate_cli_to_ir.return_value = mock_ir

    mock_response = MagicMock()
    mock_response.validation_failed = False
    mock_validate.return_value = mock_response

    mock_ctx = AsyncMock()
    mock_ctx.elicit.return_value = AcceptedElicitation(data=Consent(answer=False))

    # Execute and verify that consent was requested and error is raised
    with pytest.raises(AwsApiMcpError) as exc_info:
        await call_aws.fn('aws s3api create-bucket --bucket somebucket', mock_ctx)

    assert 'User rejected the execution of the command' in str(exc_info.value)
    mock_translate_cli_to_ir.assert_called_once_with('aws s3api create-bucket --bucket somebucket')
    mock_validate.assert_called_once_with(mock_ir)


@patch('awslabs.aws_api_mcp_server.server.DEFAULT_REGION', 'us-east-1')
@patch('awslabs.aws_api_mcp_server.server.REQUIRE_MUTATION_CONSENT', False)
@patch('awslabs.aws_api_mcp_server.server.interpret_command')
@patch('awslabs.aws_api_mcp_server.server.validate')
@patch('awslabs.aws_api_mcp_server.server.translate_cli_to_ir')
@patch('awslabs.aws_api_mcp_server.core.aws.service.is_operation_read_only')
async def test_call_aws_without_consent(
    mock_is_operation_read_only,
    mock_translate_cli_to_ir,
    mock_validate,
    mock_interpret,
):
    """Test call_aws with mutating action and with consent disabled."""
    # Create a proper ProgramInterpretationResponse mock
    mock_response = InterpretationResponse(error=None, json='{"Buckets": []}', status_code=200)

    mock_result = ProgramInterpretationResponse(
        response=mock_response,
        metadata=None,
        validation_failures=None,
        missing_context_failures=None,
        failed_constraints=None,
    )
    mock_interpret.return_value = mock_result

    mock_is_operation_read_only.return_value = False

    # Mock IR with command metadata
    mock_ir = MagicMock()
    mock_ir.command_metadata = MagicMock()
    mock_ir.command_metadata.service_sdk_name = 's3api'
    mock_ir.command_metadata.operation_sdk_name = 'create-bucket'
    mock_ir.command = MagicMock()
    mock_ir.command.is_awscli_customization = False  # Ensure interpret_command is called
    mock_ir.command.is_help_operation = False
    mock_ir.command.is_help_operation = False
    mock_translate_cli_to_ir.return_value = mock_ir

    mock_response = MagicMock()
    mock_response.validation_failed = False
    mock_validate.return_value = mock_response

    # Execute
    result = await call_aws.fn('aws s3api create-bucket --bucket somebucket', DummyCtx())

    # Verify that consent was requested
    assert result == mock_result
    mock_translate_cli_to_ir.assert_called_once_with('aws s3api create-bucket --bucket somebucket')
    mock_validate.assert_called_once_with(mock_ir)
    mock_interpret.assert_called_once()


@patch('awslabs.aws_api_mcp_server.server.translate_cli_to_ir')
async def test_call_aws_validation_error_awsmcp_error(mock_translate_cli_to_ir):
    """Test call_aws returns error details for AwsApiMcpError during validation."""
    mock_error = AwsApiMcpError('Invalid command syntax')
    mock_failure = MagicMock()
    mock_failure.reason = 'Invalid command syntax'
    mock_error.as_failure = MagicMock(return_value=mock_failure)
    mock_translate_cli_to_ir.side_effect = mock_error

    # Execute and verify
    with pytest.raises(AwsApiMcpError) as exc_info:
        await call_aws.fn('aws invalid-service invalid-operation', DummyCtx())

    assert 'Invalid command syntax' in str(exc_info.value)
    mock_translate_cli_to_ir.assert_called_once_with('aws invalid-service invalid-operation')


@patch('awslabs.aws_api_mcp_server.server.translate_cli_to_ir')
async def test_call_aws_validation_error_generic_exception(mock_translate_cli_to_ir):
    """Test call_aws returns error details for generic exception during validation."""
    mock_translate_cli_to_ir.side_effect = ValueError('Generic validation error')

    # Execute and verify
    with pytest.raises(AwsApiMcpError) as exc_info:
        await call_aws.fn('aws s3api list-buckets', DummyCtx())

    assert 'Generic validation error' in str(exc_info.value)


@patch('awslabs.aws_api_mcp_server.server.interpret_command', side_effect=NoCredentialsError())
@patch('awslabs.aws_api_mcp_server.server.validate')
@patch('awslabs.aws_api_mcp_server.server.translate_cli_to_ir')
@patch('awslabs.aws_api_mcp_server.core.aws.service.is_operation_read_only')
async def test_call_aws_no_credentials_error(
    mock_is_operation_read_only, mock_translate_cli_to_ir, mock_validate, mock_interpret
):
    """Test call_aws returns error when no AWS credentials are found."""
    # Mock IR with command metadata
    mock_ir = MagicMock()
    mock_ir.command_metadata = MagicMock()
    mock_ir.command_metadata.service_sdk_name = 's3api'
    mock_ir.command_metadata.operation_sdk_name = 'list-buckets'
    mock_ir.command = MagicMock()
    mock_ir.command.is_awscli_customization = False  # Ensure interpret_command is called
    mock_ir.command.is_help_operation = False
    mock_ir.command.is_help_operation = False
    mock_translate_cli_to_ir.return_value = mock_ir

    mock_is_operation_read_only.return_value = True

    # Mock validation response
    mock_response = MagicMock()
    mock_response.validation_failed = False
    mock_validate.return_value = mock_response

    # Execute and verify
    with pytest.raises(AwsApiMcpError) as exc_info:
        await call_aws.fn('aws s3api list-buckets', DummyCtx())

    assert 'No AWS credentials found' in str(exc_info.value)


@patch('awslabs.aws_api_mcp_server.server.DEFAULT_REGION', 'us-east-1')
@patch('awslabs.aws_api_mcp_server.server.interpret_command')
@patch('awslabs.aws_api_mcp_server.server.validate')
@patch('awslabs.aws_api_mcp_server.server.translate_cli_to_ir')
@patch('awslabs.aws_api_mcp_server.core.aws.service.is_operation_read_only')
async def test_call_aws_execution_error_awsmcp_error(
    mock_is_operation_read_only,
    mock_translate_cli_to_ir,
    mock_validate,
    mock_interpret,
):
    """Test call_aws returns error details for AwsApiMcpError during execution."""
    # Mock IR with command metadata
    mock_ir = MagicMock()
    mock_ir.command_metadata = MagicMock()
    mock_ir.command_metadata.service_sdk_name = 's3api'
    mock_ir.command_metadata.operation_sdk_name = 'list-buckets'
    mock_ir.command = MagicMock()
    mock_ir.command.is_awscli_customization = False  # Ensure interpret_command is called
    mock_ir.command.is_help_operation = False
    mock_ir.command.is_help_operation = False
    mock_translate_cli_to_ir.return_value = mock_ir

    mock_is_operation_read_only.return_value = True

    # Mock validation response
    mock_response = MagicMock()
    mock_response.validation_failed = False
    mock_validate.return_value = mock_response

    mock_error = AwsApiMcpError('Execution failed')
    mock_failure = MagicMock()
    mock_failure.reason = 'Execution failed'
    mock_error.as_failure = MagicMock(return_value=mock_failure)
    mock_interpret.side_effect = mock_error

    # Execute and verify
    with pytest.raises(AwsApiMcpError) as exc_info:
        await call_aws.fn('aws s3api list-buckets', DummyCtx())

    assert 'Execution failed' in str(exc_info.value)


@patch('awslabs.aws_api_mcp_server.server.DEFAULT_REGION', 'us-east-1')
@patch('awslabs.aws_api_mcp_server.server.interpret_command')
@patch('awslabs.aws_api_mcp_server.server.validate')
@patch('awslabs.aws_api_mcp_server.server.translate_cli_to_ir')
@patch('awslabs.aws_api_mcp_server.core.aws.service.is_operation_read_only')
async def test_call_aws_execution_error_generic_exception(
    mock_is_operation_read_only,
    mock_translate_cli_to_ir,
    mock_validate,
    mock_interpret,
):
    """Test call_aws returns error details for generic exception during execution."""
    # Mock IR with command metadata
    mock_ir = MagicMock()
    mock_ir.command_metadata = MagicMock()
    mock_ir.command_metadata.service_sdk_name = 's3api'
    mock_ir.command_metadata.operation_sdk_name = 'list-buckets'
    mock_ir.command = MagicMock()
    mock_ir.command.is_awscli_customization = False  # Ensure interpret_command is called
    mock_ir.command.is_help_operation = False
    mock_ir.command.is_help_operation = False
    mock_translate_cli_to_ir.return_value = mock_ir

    mock_is_operation_read_only.return_value = True

    # Mock validation response
    mock_response = MagicMock()
    mock_response.validation_failed = False
    mock_validate.return_value = mock_response

    mock_interpret.side_effect = RuntimeError('Generic execution error')

    # Execute and verify
    with pytest.raises(AwsApiMcpError) as exc_info:
        await call_aws.fn('aws s3api list-buckets', DummyCtx())

    assert 'Generic execution error' in str(exc_info.value)


async def test_call_aws_non_aws_command():
    """Test call_aws with command that doesn't start with 'aws'."""
    with patch(
        'awslabs.aws_api_mcp_server.server.translate_cli_to_ir'
    ) as mock_translate_cli_to_ir:
        mock_translate_cli_to_ir.side_effect = ValueError("Command must start with 'aws'")

        with pytest.raises(AwsApiMcpError) as exc_info:
            await call_aws.fn('s3api list-buckets', DummyCtx())

        assert "Command must start with 'aws'" in str(exc_info.value)


@patch('awslabs.aws_api_mcp_server.server.validate')
@patch('awslabs.aws_api_mcp_server.server.translate_cli_to_ir')
@patch('awslabs.aws_api_mcp_server.core.aws.service.is_operation_read_only')
@patch('awslabs.aws_api_mcp_server.server.READ_OPERATIONS_ONLY_MODE')
async def test_when_operation_is_not_allowed(
    mock_read_operations_only_mode,
    mock_is_operation_read_only,
    mock_translate_cli_to_ir,
    mock_validate,
):
    """Test call_aws returns error when operation is not allowed in read-only mode."""
    # Mock IR with command metadata
    mock_ir = MagicMock()
    mock_ir.command_metadata = MagicMock()
    mock_ir.command_metadata.service_sdk_name = 's3api'
    mock_ir.command_metadata.operation_sdk_name = 'list-buckets'
    mock_ir.command = MagicMock()
    mock_ir.command.is_awscli_customization = False  # Ensure interpret_command is called
    mock_ir.command.is_help_operation = False
    mock_ir.command.is_help_operation = False
    mock_translate_cli_to_ir.return_value = mock_ir

    mock_read_operations_only_mode.return_value = True

    # Mock validation response
    mock_response = MagicMock()
    mock_response.validation_failed = False
    mock_validate.return_value = mock_response

    mock_is_operation_read_only.return_value = False

    # Execute and verify
    with pytest.raises(AwsApiMcpError) as exc_info:
        await call_aws.fn('aws s3api list-buckets', DummyCtx())

    assert 'Execution of this operation is not allowed because read only mode is enabled' in str(
        exc_info.value
    )


@patch('awslabs.aws_api_mcp_server.server.validate')
@patch('awslabs.aws_api_mcp_server.server.translate_cli_to_ir')
async def test_call_aws_validation_failures(mock_translate_cli_to_ir, mock_validate):
    """Test call_aws returns error for validation failures."""
    # Mock IR with command metadata
    mock_ir = MagicMock()
    mock_ir.command_metadata = MagicMock()
    mock_ir.command_metadata.service_sdk_name = 's3api'
    mock_ir.command_metadata.operation_sdk_name = 'list-buckets'
    mock_ir.command = MagicMock()
    mock_ir.command.is_awscli_customization = False  # Ensure interpret_command is called
    mock_ir.command.is_help_operation = False
    mock_ir.command.is_help_operation = False
    mock_translate_cli_to_ir.return_value = mock_ir

    # Mock validation response with validation failures
    mock_response = MagicMock()
    mock_response.validation_failures = ['Invalid parameter value']
    mock_response.failed_constraints = None
    mock_response.model_dump_json.return_value = (
        '{"validation_failures": ["Invalid parameter value"]}'
    )
    mock_validate.return_value = mock_response

    # Execute and verify
    with pytest.raises(CommandValidationError) as exc_info:
        await call_aws.fn('aws s3api list-buckets', DummyCtx())

    assert 'Invalid parameter value' in str(exc_info.value)
    mock_translate_cli_to_ir.assert_called_once_with('aws s3api list-buckets')
    mock_validate.assert_called_once_with(mock_ir)


@patch('awslabs.aws_api_mcp_server.server.validate')
@patch('awslabs.aws_api_mcp_server.server.translate_cli_to_ir')
async def test_call_aws_failed_constraints(mock_translate_cli_to_ir, mock_validate):
    """Test call_aws returns error for failed constraints."""
    # Mock IR with command metadata
    mock_ir = MagicMock()
    mock_ir.command_metadata = MagicMock()
    mock_ir.command_metadata.service_sdk_name = 's3api'
    mock_ir.command_metadata.operation_sdk_name = 'list-buckets'
    mock_ir.command = MagicMock()
    mock_ir.command.is_awscli_customization = False  # Ensure interpret_command is called
    mock_ir.command.is_help_operation = False
    mock_ir.command.is_help_operation = False
    mock_translate_cli_to_ir.return_value = mock_ir

    # Mock validation response with failed constraints
    mock_response = MagicMock()
    mock_response.validation_failures = None
    mock_response.failed_constraints = ['Resource limit exceeded']
    mock_response.model_dump_json.return_value = (
        '{"failed_constraints": ["Resource limit exceeded"]}'
    )
    mock_validate.return_value = mock_response

    # Execute and verify
    with pytest.raises(CommandValidationError) as exc_info:
        await call_aws.fn('aws s3api list-buckets', DummyCtx())

    assert 'Resource limit exceeded' in str(exc_info.value)
    mock_translate_cli_to_ir.assert_called_once_with('aws s3api list-buckets')
    mock_validate.assert_called_once_with(mock_ir)


@patch('awslabs.aws_api_mcp_server.server.validate')
@patch('awslabs.aws_api_mcp_server.server.translate_cli_to_ir')
async def test_call_aws_both_validation_failures_and_constraints(
    mock_translate_cli_to_ir, mock_validate
):
    """Test call_aws returns error for both validation failures and failed constraints."""
    # Mock IR with command metadata
    mock_ir = MagicMock()
    mock_ir.command_metadata = MagicMock()
    mock_ir.command_metadata.service_sdk_name = 's3api'
    mock_ir.command_metadata.operation_sdk_name = 'list-buckets'
    mock_ir.command = MagicMock()
    mock_ir.command.is_awscli_customization = False  # Ensure interpret_command is called
    mock_ir.command.is_help_operation = False
    mock_ir.command.is_help_operation = False
    mock_translate_cli_to_ir.return_value = mock_ir

    # Mock validation response with both validation failures and failed constraints
    mock_response = MagicMock()
    mock_response.validation_failures = ['Invalid parameter value']
    mock_response.failed_constraints = ['Resource limit exceeded']
    mock_response.model_dump_json.return_value = '{"validation_failures": ["Invalid parameter value"], "failed_constraints": ["Resource limit exceeded"]}'
    mock_validate.return_value = mock_response

    # Execute and verify
    with pytest.raises(CommandValidationError) as exc_info:
        await call_aws.fn('aws s3api list-buckets', DummyCtx())

    error_msg = str(exc_info.value)
    assert 'Invalid parameter value' in error_msg
    assert 'Resource limit exceeded' in error_msg
    mock_translate_cli_to_ir.assert_called_once_with('aws s3api list-buckets')
    mock_validate.assert_called_once_with(mock_ir)


@patch('awslabs.aws_api_mcp_server.server.execute_awscli_customization')
@patch('awslabs.aws_api_mcp_server.server.validate')
@patch('awslabs.aws_api_mcp_server.server.translate_cli_to_ir')
@patch('awslabs.aws_api_mcp_server.core.aws.service.is_operation_read_only')
async def test_call_aws_awscli_customization_success(
    mock_is_operation_read_only,
    mock_translate_cli_to_ir,
    mock_validate,
    mock_execute_awscli_customization,
):
    """Test call_aws returns success response for AWS CLI customization command."""
    mock_ir = MagicMock()
    mock_ir.command = MagicMock()
    mock_ir.command.is_awscli_customization = True
    mock_ir.command.is_help_operation = False
    mock_translate_cli_to_ir.return_value = mock_ir

    mock_is_operation_read_only.return_value = True

    mock_response = MagicMock()
    mock_response.validation_failed = False
    mock_validate.return_value = mock_response

    expected_response = AwsCliAliasResponse(response='Command executed successfully', error=None)
    mock_execute_awscli_customization.return_value = expected_response

    result = await call_aws.fn('aws configure list', DummyCtx())

    assert result == expected_response
    mock_translate_cli_to_ir.assert_called_once_with('aws configure list')
    mock_validate.assert_called_once_with(mock_ir)
    mock_execute_awscli_customization.assert_called_once_with(
        'aws configure list',
        mock_ir.command,
        credentials=None,
        default_region_override=None,
    )


@patch('awslabs.aws_api_mcp_server.server.execute_awscli_customization')
@patch('awslabs.aws_api_mcp_server.server.validate')
@patch('awslabs.aws_api_mcp_server.server.translate_cli_to_ir')
@patch('awslabs.aws_api_mcp_server.core.aws.service.is_operation_read_only')
async def test_call_aws_awscli_customization_error(
    mock_is_operation_read_only,
    mock_translate_cli_to_ir,
    mock_validate,
    mock_execute_awscli_customization,
):
    """Test call_aws handles error response from AWS CLI customization command."""
    mock_ir = MagicMock()
    mock_ir.command = MagicMock()
    mock_ir.command.is_awscli_customization = True
    mock_ir.command.is_help_operation = False
    mock_translate_cli_to_ir.return_value = mock_ir

    mock_is_operation_read_only.return_value = True

    mock_response = MagicMock()
    mock_response.validation_failed = False
    mock_validate.return_value = mock_response

    mock_execute_awscli_customization.side_effect = AwsApiMcpError(
        "Error while executing 'aws configure list': Configuration file not found"
    )

    with pytest.raises(AwsApiMcpError) as exc_info:
        await call_aws.fn('aws configure list', DummyCtx())

    assert 'Configuration file not found' in str(exc_info.value)
    mock_translate_cli_to_ir.assert_called_once_with('aws configure list')
    mock_validate.assert_called_once_with(mock_ir)
    mock_execute_awscli_customization.assert_called_once_with(
        'aws configure list',
        mock_ir.command,
        credentials=None,
        default_region_override=None,
    )


@patch('awslabs.aws_api_mcp_server.server.DEFAULT_REGION', None)
@patch('awslabs.aws_api_mcp_server.server.WORKING_DIRECTORY', '/tmp')
def test_main_missing_aws_region():
    """Test main function raises ValueError when AWS_REGION environment variable is not set."""
    with pytest.raises(ValueError, match=r'AWS_REGION environment variable is not defined.'):
        main()


@patch('awslabs.aws_api_mcp_server.server.DEFAULT_REGION', 'us-east-1')
@patch('awslabs.aws_api_mcp_server.server.WORKING_DIRECTORY', 'relative/path')
def test_main_relative_working_directory():
    """Test main function raises ValueError when AWS_API_MCP_WORKING_DIR is a relative path."""
    with pytest.raises(
        ValueError,
        match=r'AWS_API_MCP_WORKING_DIR must be an absolute path.',
    ):
        main()


@patch('awslabs.aws_api_mcp_server.server.os.chdir')
@patch('awslabs.aws_api_mcp_server.server.server')
@patch('awslabs.aws_api_mcp_server.server.get_read_only_operations')
@patch('awslabs.aws_api_mcp_server.server.READ_OPERATIONS_ONLY_MODE', True)
@patch('awslabs.aws_api_mcp_server.server.DEFAULT_REGION', 'us-east-1')
@patch('awslabs.aws_api_mcp_server.server.WORKING_DIRECTORY', '/tmp')
def test_main_success_with_read_only_mode(
    mock_get_read_only_operations,
    mock_server,
    mock_chdir,
):
    """Test main function executes successfully with read-only mode enabled."""
    mock_read_operations = MagicMock()
    mock_get_read_only_operations.return_value = mock_read_operations
    mock_server.run = MagicMock()

    main()

    mock_chdir.assert_called_once_with('/tmp')
    mock_get_read_only_operations.assert_called_once()
    mock_server.run.assert_called_once_with(transport='stdio')


@patch('awslabs.aws_api_mcp_server.server.os.chdir')
@patch('awslabs.aws_api_mcp_server.server.server')
@patch('awslabs.aws_api_mcp_server.server.get_read_only_operations')
@patch('awslabs.aws_api_mcp_server.server.READ_OPERATIONS_ONLY_MODE', True)
@patch('awslabs.aws_api_mcp_server.server.DEFAULT_REGION', 'us-east-1')
@patch('awslabs.aws_api_mcp_server.server.WORKING_DIRECTORY', '/tmp')
@patch('awslabs.aws_api_mcp_server.server.TRANSPORT', 'streamable-http')
@patch('awslabs.aws_api_mcp_server.server.HOST', '0.0.0.0')
@patch('awslabs.aws_api_mcp_server.server.PORT', 8080)
@patch('awslabs.aws_api_mcp_server.server.STATELESS_HTTP', True)
def test_main_success_with_http_transport(
    mock_get_read_only_operations,
    mock_server,
    mock_chdir,
):
    """Test main function executes successfully with HTTP transport (else branch)."""
    mock_read_operations = MagicMock()
    mock_get_read_only_operations.return_value = mock_read_operations
    mock_server.run = MagicMock()

    main()

    mock_chdir.assert_called_once_with('/tmp')
    mock_get_read_only_operations.assert_called_once()
    mock_server.run.assert_called_once_with(
        transport='streamable-http',
        host='0.0.0.0',
        port=8080,
        stateless_http=True,
    )


@patch('awslabs.aws_api_mcp_server.core.common.config.ENABLE_AGENT_SCRIPTS', True)
async def test_get_execution_plan_is_available_when_env_var_is_set():
    """Test get_execution_plan returns script content when script exists."""
    # Re-import the server module to ensure the tool is registered
    import awslabs.aws_api_mcp_server.server
    import importlib

    importlib.reload(awslabs.aws_api_mcp_server.server)

    from awslabs.aws_api_mcp_server.server import server

    tools = await server._list_tools_middleware()
    tool_names = [tool.name for tool in tools]
    assert 'get_execution_plan' in tool_names


@patch('awslabs.aws_api_mcp_server.core.common.config.ENABLE_AGENT_SCRIPTS', False)
async def test_get_execution_plan_is_available_when_env_var_is_not_set():
    """Test get_execution_plan returns script content when script exists."""
    # Re-import the server module to ensure the tool is not registered
    import awslabs.aws_api_mcp_server.server
    import importlib

    importlib.reload(awslabs.aws_api_mcp_server.server)

    from awslabs.aws_api_mcp_server.server import server

    tools = await server._list_tools_middleware()
    tool_names = [tool.name for tool in tools]
    assert 'get_execution_plan' not in tool_names


@patch('awslabs.aws_api_mcp_server.core.common.config.ENABLE_AGENT_SCRIPTS', True)
async def test_get_execution_plan_script_not_found():
    """Test get_execution_plan returns error when script does not exist."""
    # Re-import the server module to ensure the function is defined
    import awslabs.aws_api_mcp_server.server
    import importlib

    importlib.reload(awslabs.aws_api_mcp_server.server)

    from awslabs.aws_api_mcp_server.server import get_execution_plan

    # Mock the AGENT_SCRIPTS_MANAGER after reloading
    with patch(
        'awslabs.aws_api_mcp_server.server.AGENT_SCRIPTS_MANAGER'
    ) as mock_agent_scripts_manager:
        mock_agent_scripts_manager.get_script.return_value = None

        with pytest.raises(AwsApiMcpError) as exc_info:
            await get_execution_plan.fn('non-existent-script', DummyCtx())

        assert 'Script non-existent-script not found' in str(exc_info.value)
        mock_agent_scripts_manager.get_script.assert_called_once_with('non-existent-script')


@patch('awslabs.aws_api_mcp_server.core.common.config.ENABLE_AGENT_SCRIPTS', True)
async def test_get_execution_plan_exception_handling():
    """Test get_execution_plan handles exceptions properly."""
    # Re-import the server module to ensure the function is defined
    import awslabs.aws_api_mcp_server.server
    import importlib

    importlib.reload(awslabs.aws_api_mcp_server.server)

    from awslabs.aws_api_mcp_server.server import get_execution_plan

    # Mock the AGENT_SCRIPTS_MANAGER after reloading
    with patch(
        'awslabs.aws_api_mcp_server.server.AGENT_SCRIPTS_MANAGER'
    ) as mock_agent_scripts_manager:
        mock_agent_scripts_manager.get_script.side_effect = Exception('Test exception')

        with pytest.raises(AwsApiMcpError) as exc_info:
            await get_execution_plan.fn('test-script', DummyCtx())

        assert 'Test exception' in str(exc_info.value)


# Tests for call_aws_helper function
@patch('awslabs.aws_api_mcp_server.server.interpret_command')
@patch('awslabs.aws_api_mcp_server.server.validate')
@patch('awslabs.aws_api_mcp_server.server.translate_cli_to_ir')
async def test_call_aws_helper_with_credentials(mock_translate, mock_validate, mock_interpret):
    """Test call_aws_helper passes credentials to interpret_command."""
    test_credentials = Credentials(**TEST_CREDENTIALS)

    # Mock IR with command metadata
    mock_ir = MagicMock()
    mock_ir.command_metadata = MagicMock()
    mock_ir.command_metadata.service_sdk_name = 's3api'
    mock_ir.command_metadata.operation_sdk_name = 'list-buckets'
    mock_ir.command.is_awscli_customization = False  # Ensure interpret_command is called
    mock_ir.command.is_help_operation = False
    mock_translate.return_value = mock_ir

    mock_validation = MagicMock()
    mock_validation.validation_failed = False
    mock_validate.return_value = mock_validation

    mock_response = MagicMock()
    mock_interpret.return_value = mock_response

    result = await call_aws_helper(
        'aws s3api list-buckets',
        AsyncMock(),
        credentials=test_credentials,
    )

    mock_interpret.assert_called_once_with(
        cli_command='aws s3api list-buckets',
        max_results=None,
        credentials=test_credentials,
        default_region_override=None,
    )
    assert result == mock_response


@patch('awslabs.aws_api_mcp_server.server.interpret_command')
@patch('awslabs.aws_api_mcp_server.server.validate')
@patch('awslabs.aws_api_mcp_server.server.translate_cli_to_ir')
async def test_call_aws_helper_without_credentials(mock_translate, mock_validate, mock_interpret):
    """Test call_aws_helper works without credentials."""
    # Mock IR with command metadata
    mock_ir = MagicMock()
    mock_ir.command_metadata = MagicMock()
    mock_ir.command_metadata.service_sdk_name = 's3api'
    mock_ir.command_metadata.operation_sdk_name = 'list-buckets'
    mock_ir.command.is_awscli_customization = False  # Ensure interpret_command is called
    mock_ir.command.is_help_operation = False
    mock_translate.return_value = mock_ir

    mock_validation = MagicMock()
    mock_validation.validation_failed = False
    mock_validate.return_value = mock_validation

    mock_response = MagicMock()
    mock_interpret.return_value = mock_response

    result = await call_aws_helper(
        'aws s3api list-buckets',
        AsyncMock(),
        credentials=None,
    )

    mock_interpret.assert_called_once_with(
        cli_command='aws s3api list-buckets',
        max_results=None,
        credentials=None,
        default_region_override=None,
    )
    assert result == mock_response


@patch('awslabs.aws_api_mcp_server.server.call_aws_helper')
async def test_call_aws_delegates_to_helper(mock_call_aws_helper):
    """Test call_aws delegates to call_aws_helper with None credentials."""
    mock_response = MagicMock()
    mock_call_aws_helper.return_value = mock_response

    ctx = DummyCtx()

    result = await call_aws.fn('aws s3api list-buckets', ctx)

    mock_call_aws_helper.assert_called_once_with(
        cli_command='aws s3api list-buckets', ctx=ctx, max_results=None, credentials=None
    )
    assert result == mock_response


@pytest.mark.parametrize(
    'service, operation',
    [
        ('s3', 'ls'),
        ('s3api', 'list-buckets'),
        ('ec2', 'describe-instances'),
        ('lambda', 'list-functions'),
        ('dynamodb', 'list-tables'),
        ('iam', 'get-user'),
        ('cloudwatch', 'get-metric-data'),
        ('apigateway', 'get-rest-apis'),
        ('rds', 'describe-db-instances'),
        ('sns', 'list-topics'),
        ('sqs', 'list-queues'),
        ('sts', 'get-caller-identity'),
        ('cloudformation', 'describe-stacks'),
        ('kms', 'list-keys'),
        ('elasticbeanstalk', 'describe-environments'),
        ('organizations', 'list-accounts'),
        ('ec2', 'describe-volumes'),
        ('ecs', 'list-clusters'),
        ('efs', 'describe-file-systems'),
        ('route53', 'list-hosted-zones'),
        ('lightsail', 'get-instances'),
    ],
)
async def test_call_aws_help_command_success(service, operation):
    """Test call_aws returns success response for help command."""
    help_document = generate_help_document(service, operation)
    assert help_document is not None
    expected_response = ProgramInterpretationResponse(
        response=InterpretationResponse(error=None, json=as_json(help_document), status_code=200),
        metadata=None,
        validation_failures=None,
        missing_context_failures=None,
        failed_constraints=None,
    )

    result = await call_aws.fn(f'aws {service} {operation} help', DummyCtx())

    assert result == expected_response


@patch('awslabs.aws_api_mcp_server.server.get_help_document')
@patch('awslabs.aws_api_mcp_server.server.validate')
@patch('awslabs.aws_api_mcp_server.server.translate_cli_to_ir')
async def test_call_aws_help_command_failure(
    mock_translate_cli_to_ir,
    mock_validate,
    mock_get_help_document,
):
    """Test call_aws raises error when help command fails."""
    mock_ir = MagicMock()
    mock_ir.command = MagicMock()
    mock_ir.command.is_awscli_customization = False
    mock_ir.command.is_help_operation = True
    mock_translate_cli_to_ir.return_value = mock_ir

    mock_response = MagicMock()
    mock_response.validation_failed = False
    mock_validate.return_value = mock_response

    mock_get_help_document.side_effect = AwsApiMcpError('Failed to generate help document')

    with pytest.raises(AwsApiMcpError) as exc_info:
        await call_aws.fn('aws non-existing-service non-existing-operation help', DummyCtx())

    assert 'Failed to generate help document' in str(exc_info.value)
    mock_translate_cli_to_ir.assert_called_once_with(
        'aws non-existing-service non-existing-operation help'
    )
    mock_validate.assert_called_once_with(mock_ir)
    mock_get_help_document.assert_called_once()



import pytest
from awslabs.aws_api_mcp_server.middleware.http_header_validation_middleware import (
    HTTPHeaderValidationMiddleware,
)
from fastmcp.exceptions import ClientError
from unittest.mock import AsyncMock, MagicMock, patch


@pytest.mark.parametrize(
    'origin_value,allowed_origins',
    [
        ('http://example.com', 'http://example.com'),  # Exact match
        ('https://example.com:3000', 'https://example.com'),  # With port
        ('http://example.com', 'http://example.com,http://other.com'),  # Multiple allowed origins
        ('http://other.com', 'http://example.com,http://other.com'),  # Second in list
        ('https://example.com', '*'),  # Wildcard
        ('http://any-domain.com', '*'),  # Wildcard allows any
    ],
)
@patch('awslabs.aws_api_mcp_server.middleware.http_header_validation_middleware.get_http_headers')
@pytest.mark.asyncio
async def test_origin_header_validation_passes(
    mock_get_headers: MagicMock,
    origin_value: str,
    allowed_origins: str,
):
    """Test origin header validation passes for allowed origins."""
    mock_get_headers.return_value = {'origin': origin_value}

    middleware = HTTPHeaderValidationMiddleware()
    context = MagicMock()
    call_next = AsyncMock(return_value='success')

    with patch(
        'awslabs.aws_api_mcp_server.middleware.http_header_validation_middleware.ALLOWED_ORIGINS',
        allowed_origins,
    ):
        result = await middleware.on_request(context, call_next)
        assert result == 'success'
        call_next.assert_called_once_with(context)


@pytest.mark.parametrize(
    'origin_value,allowed_origins',
    [
        ('http://forbidden.com', 'http://example.com'),  # Not in allowed list
        ('http://forbidden.com', 'http://example.com,http://other.com'),  # Not in multiple allowed
        ('http://sub.example.com', 'http://example.com'),  # Subdomain not matched
    ],
)
@patch('awslabs.aws_api_mcp_server.middleware.http_header_validation_middleware.get_http_headers')
@pytest.mark.asyncio
async def test_origin_header_validation_fails(
    mock_get_headers: MagicMock,
    origin_value: str,
    allowed_origins: str,
):
    """Test origin header validation fails for disallowed origins."""
    mock_get_headers.return_value = {'origin': origin_value}

    middleware = HTTPHeaderValidationMiddleware()
    context = MagicMock()
    call_next = AsyncMock(return_value='success')

    with patch(
        'awslabs.aws_api_mcp_server.middleware.http_header_validation_middleware.ALLOWED_ORIGINS',
        allowed_origins,
    ):
        with pytest.raises(ClientError, match='Origin header validation failed'):
            await middleware.on_request(context, call_next)
        call_next.assert_not_called()


@pytest.mark.parametrize(
    'host_value,allowed_hosts',
    [
        ('example.com', 'example.com'),  # Exact match
        ('example.com:8080', 'example.com'),  # With port
        ('example.com', 'example.com,other.com'),  # Multiple allowed hosts
        ('other.com', 'example.com,other.com'),  # Second in list
        ('example.com', '*'),  # Wildcard
        ('any-domain.com', '*'),  # Wildcard allows any
        ('127.0.0.1', '127.0.0.1'),  # IP address
        ('localhost:3000', 'localhost'),  # localhost with port
    ],
)
@patch('awslabs.aws_api_mcp_server.middleware.http_header_validation_middleware.get_http_headers')
@pytest.mark.asyncio
async def test_host_header_validation_passes(
    mock_get_headers: MagicMock,
    host_value: str,
    allowed_hosts: str,
):
    """Test host header validation passes for allowed hosts."""
    # No origin header, only host
    mock_get_headers.return_value = {'host': host_value}

    middleware = HTTPHeaderValidationMiddleware()
    context = MagicMock()
    call_next = AsyncMock(return_value='success')

    with patch(
        'awslabs.aws_api_mcp_server.middleware.http_header_validation_middleware.ALLOWED_HOSTS',
        allowed_hosts,
    ):
        result = await middleware.on_request(context, call_next)
        assert result == 'success'
        call_next.assert_called_once_with(context)


@pytest.mark.parametrize(
    'host_value,allowed_hosts',
    [
        ('forbidden.com', 'example.com'),  # Not in allowed list
        ('malicious.com', '127.0.0.1'),
        ('other.com:8080', 'example.com'),
        ('forbidden.com', 'example.com,other.com'),  # Not in multiple allowed
        ('sub.example.com', 'example.com'),  # Subdomain not matched
    ],
)
@patch('awslabs.aws_api_mcp_server.middleware.http_header_validation_middleware.get_http_headers')
@pytest.mark.asyncio
async def test_host_header_validation_fails(
    mock_get_headers: MagicMock,
    host_value: str,
    allowed_hosts: str,
):
    """Test host header validation fails for disallowed hosts."""
    # No origin header, only host
    mock_get_headers.return_value = {'host': host_value}

    middleware = HTTPHeaderValidationMiddleware()
    context = MagicMock()
    call_next = AsyncMock(return_value='success')

    with patch(
        'awslabs.aws_api_mcp_server.middleware.http_header_validation_middleware.ALLOWED_HOSTS',
        allowed_hosts,
    ):
        with pytest.raises(ClientError, match='Host header validation failed'):
            await middleware.on_request(context, call_next)
        call_next.assert_not_called()


@patch('awslabs.aws_api_mcp_server.middleware.http_header_validation_middleware.get_http_headers')
@pytest.mark.asyncio
async def test_both_headers_validated_independently(mock_get_headers: MagicMock):
    """Test that both host and origin headers are validated independently."""
    # Both headers present
    mock_get_headers.return_value = {
        'origin': 'http://example.com',
        'host': 'example.com',
    }

    middleware = HTTPHeaderValidationMiddleware()
    context = MagicMock()
    call_next = AsyncMock(return_value='success')

    with (
        patch(
            'awslabs.aws_api_mcp_server.middleware.http_header_validation_middleware.ALLOWED_ORIGINS',
            'http://example.com',
        ),
        patch(
            'awslabs.aws_api_mcp_server.middleware.http_header_validation_middleware.ALLOWED_HOSTS',
            'example.com',
        ),
    ):
        # Both should pass validation
        result = await middleware.on_request(context, call_next)
        assert result == 'success'
        call_next.assert_called_once_with(context)


@patch('awslabs.aws_api_mcp_server.middleware.http_header_validation_middleware.get_http_headers')
@pytest.mark.asyncio
async def test_host_fails_validation_when_both_present(mock_get_headers: MagicMock):
    """Test that host validation fails even when origin is valid."""
    # Both headers present, origin valid but host invalid
    mock_get_headers.return_value = {
        'origin': 'http://example.com',
        'host': 'malicious.com',
    }

    middleware = HTTPHeaderValidationMiddleware()
    context = MagicMock()
    call_next = AsyncMock(return_value='success')

    with (
        patch(
            'awslabs.aws_api_mcp_server.middleware.http_header_validation_middleware.ALLOWED_ORIGINS',
            'http://example.com',
        ),
        patch(
            'awslabs.aws_api_mcp_server.middleware.http_header_validation_middleware.ALLOWED_HOSTS',
            'example.com',
        ),
    ):
        # Should fail on host validation
        with pytest.raises(ClientError, match='Host header validation failed'):
            await middleware.on_request(context, call_next)
        call_next.assert_not_called()


@patch('awslabs.aws_api_mcp_server.middleware.http_header_validation_middleware.get_http_headers')
@pytest.mark.asyncio
async def test_origin_fails_validation_when_both_present(mock_get_headers: MagicMock):
    """Test that origin validation fails even when host is valid."""
    # Both headers present, host valid but origin invalid
    mock_get_headers.return_value = {
        'origin': 'http://malicious.com',
        'host': 'example.com',
    }

    middleware = HTTPHeaderValidationMiddleware()
    context = MagicMock()
    call_next = AsyncMock(return_value='success')

    with (
        patch(
            'awslabs.aws_api_mcp_server.middleware.http_header_validation_middleware.ALLOWED_ORIGINS',
            'http://example.com',
        ),
        patch(
            'awslabs.aws_api_mcp_server.middleware.http_header_validation_middleware.ALLOWED_HOSTS',
            'example.com',
        ),
    ):
        # Should fail on origin validation
        with pytest.raises(ClientError, match='Origin header validation failed'):
            await middleware.on_request(context, call_next)
        call_next.assert_not_called()


@patch('awslabs.aws_api_mcp_server.middleware.http_header_validation_middleware.get_http_headers')
@pytest.mark.asyncio
async def test_no_origin_or_host_headers(mock_get_headers: MagicMock):
    """Test that request passes through when neither origin nor host headers are present."""
    mock_get_headers.return_value = {}

    middleware = HTTPHeaderValidationMiddleware()
    context = MagicMock()
    call_next = AsyncMock(return_value='success')

    result = await middleware.on_request(context, call_next)
    assert result == 'success'
    call_next.assert_called_once_with(context)


@pytest.mark.parametrize(
    'origin_with_port,expected_origin',
    [
        ('http://example.com:3000', 'http://example.com'),
        ('https://example.com:8080', 'https://example.com'),
        ('http://localhost:5000', 'http://localhost'),
        ('http://192.168.1.1:8000', 'http://192.168.1.1'),
        ('https://example.com', 'https://example.com'),
    ],
)
@patch('awslabs.aws_api_mcp_server.middleware.http_header_validation_middleware.get_http_headers')
@pytest.mark.asyncio
async def test_port_removal_from_origin(
    mock_get_headers: MagicMock,
    origin_with_port: str,
    expected_origin: str,
):
    """Test that port is correctly removed from origin before validation."""
    mock_get_headers.return_value = {'origin': origin_with_port}

    middleware = HTTPHeaderValidationMiddleware()
    context = MagicMock()
    call_next = AsyncMock(return_value='success')

    with patch(
        'awslabs.aws_api_mcp_server.middleware.http_header_validation_middleware.ALLOWED_ORIGINS',
        expected_origin,
    ):
        result = await middleware.on_request(context, call_next)
        assert result == 'success'
        call_next.assert_called_once_with(context)


@patch('awslabs.aws_api_mcp_server.middleware.http_header_validation_middleware.get_http_headers')
@pytest.mark.asyncio
async def test_empty_allowed_origins(mock_get_headers: MagicMock):
    """Test behavior when ALLOWED_ORIGINS is empty."""
    mock_get_headers.return_value = {'origin': 'http://example.com'}

    middleware = HTTPHeaderValidationMiddleware()
    context = MagicMock()
    call_next = AsyncMock()

    with patch(
        'awslabs.aws_api_mcp_server.middleware.http_header_validation_middleware.ALLOWED_ORIGINS',
        '',
    ):
        # Should fail validation with empty allowed origins
        with pytest.raises(ClientError, match='Origin header validation failed'):
            await middleware.on_request(context, call_next)



import awslabs.aws_api_mcp_server.core.common.config as config_module
import importlib
import pytest
from awslabs.aws_api_mcp_server.core.common.config import (
    FileAccessMode,
    get_file_access_mode,
    get_region,
    get_server_directory,
    get_transport_from_env,
    get_user_agent_extra,
)
from importlib.metadata import PackageNotFoundError
from unittest.mock import MagicMock, patch


@pytest.mark.parametrize(
    'os_name,uname_sysname,expected_tempdir',
    [
        ('nt', 'Windows', '/tmp'),
        ('posix', 'Darwin', '/private/var/folders/rq/'),
    ],
)
@patch('awslabs.aws_api_mcp_server.core.common.config.Path')
@patch('tempfile.gettempdir')
@patch('os.uname')
@patch('os.name')
def test_get_server_directory_windows_macos(
    mock_os_name: MagicMock,
    mock_uname: MagicMock,
    mock_tempdir: MagicMock,
    mock_path: MagicMock,
    os_name: str,
    uname_sysname: str,
    expected_tempdir: str,
):
    """Test get_server_directory for Windows and macOS platforms."""
    mock_os_name.return_value = os_name
    mock_uname.return_value.sysname = uname_sysname
    mock_tempdir.return_value = expected_tempdir

    mock_path_instance = MagicMock()
    mock_path_instance.__truediv__ = lambda self, other: f'{expected_tempdir}/{other}'
    mock_path_instance.__str__ = lambda: expected_tempdir
    mock_path.return_value = mock_path_instance

    result = get_server_directory()

    assert f'{expected_tempdir}/aws-api-mcp' in str(result)


@pytest.mark.parametrize(
    'aws_region,profile_name,profile_region,default_region,expected_region',
    [
        (
            'us-west-1',
            'profile1',
            'eu-west-1',
            'ap-south-1',
            'us-west-1',
        ),  # AWS_REGION takes precedence
        (None, 'profile1', 'eu-west-1', 'ap-south-1', 'eu-west-1'),  # Profile region used
        (None, 'profile1', None, 'ap-south-1', 'us-east-1'),  # Profile has no region, fallback
        (None, None, None, 'ap-south-1', 'ap-south-1'),  # Default session region used
        (None, None, None, None, 'us-east-1'),  # All None, fallback used
    ],
)
@patch('awslabs.aws_api_mcp_server.core.common.config.boto3.Session')
def test_get_region_parametrized(
    mock_session_class: MagicMock,
    aws_region: str,
    profile_name: str,
    profile_region: str,
    default_region: str,
    expected_region: str,
):
    """Parametrized test for various combinations of region sources."""
    profile_session = MagicMock()
    profile_session.region_name = profile_region

    default_session = MagicMock()
    default_session.region_name = default_region

    # Configure mock_session_class to return different sessions based on arguments
    def session_side_effect(*args, **kwargs):
        if 'profile_name' in kwargs:
            return profile_session
        return default_session

    mock_session_class.side_effect = session_side_effect

    with patch('awslabs.aws_api_mcp_server.core.common.config.AWS_REGION', aws_region):
        result = get_region(profile_name)

    assert result == expected_region


@pytest.mark.parametrize(
    'transport_value,expected',
    [
        ('stdio', 'stdio'),
        ('streamable-http', 'streamable-http'),
    ],
)
def test_get_transport_from_env_valid_values(monkeypatch, transport_value, expected):
    """Test get_transport_from_env with valid transport values."""
    monkeypatch.setenv('AWS_API_MCP_TRANSPORT', transport_value)

    # For streamable-http, AUTH_TYPE must be explicitly set to 'no-auth'
    if transport_value == 'streamable-http':
        monkeypatch.setenv('AUTH_TYPE', 'no-auth')
    else:
        monkeypatch.delenv('AUTH_TYPE', raising=False)

    result = get_transport_from_env()

    assert result == expected


def test_get_transport_from_env_default_value(monkeypatch):
    """Test get_transport_from_env returns default value when env var is not set."""
    # Ensure the environment variable is not set
    monkeypatch.delenv('AWS_API_MCP_TRANSPORT', raising=False)
    monkeypatch.delenv('AUTH_TYPE', raising=False)

    result = get_transport_from_env()

    assert result == 'stdio'


@pytest.mark.parametrize(
    'invalid_transport',
    [
        'http',
        'websocket',
        'tcp',
        'invalid',
        'STDIO',
        'STREAMABLE-HTTP',
        '',
        'stdio-http',
    ],
)
def test_get_transport_from_env_invalid_values(monkeypatch, invalid_transport):
    """Test get_transport_from_env raises ValueError for invalid transport values."""
    monkeypatch.setenv('AWS_API_MCP_TRANSPORT', invalid_transport)
    monkeypatch.delenv('AUTH_TYPE', raising=False)

    with pytest.raises(ValueError, match=f'Invalid transport: {invalid_transport}'):
        get_transport_from_env()


def test_get_transport_from_env_streamable_http_requires_auth_type(monkeypatch):
    """Ensure streamable-http transport fails without explicit AUTH_TYPE=no-auth."""
    monkeypatch.setenv('AWS_API_MCP_TRANSPORT', 'streamable-http')
    monkeypatch.delenv('AUTH_TYPE', raising=False)

    with pytest.raises(
        ValueError,
        match="requires AUTH_TYPE environment variable to be explicitly set to 'no-auth'",
    ):
        get_transport_from_env()


def test_get_transport_from_env_streamable_http_with_wrong_auth_type(monkeypatch):
    """Ensure streamable-http transport fails when AUTH_TYPE is not no-auth."""
    monkeypatch.setenv('AWS_API_MCP_TRANSPORT', 'streamable-http')
    monkeypatch.setenv('AUTH_TYPE', 'basic')

    with pytest.raises(
        ValueError,
        match="requires AUTH_TYPE environment variable to be explicitly set to 'no-auth'",
    ):
        get_transport_from_env()


def test_get_transport_from_env_streamable_http_with_no_auth(monkeypatch):
    """Ensure streamable-http transport succeeds when AUTH_TYPE=no-auth."""
    monkeypatch.setenv('AWS_API_MCP_TRANSPORT', 'streamable-http')
    monkeypatch.setenv('AUTH_TYPE', 'no-auth')

    assert get_transport_from_env() == 'streamable-http'


@patch('awslabs.aws_api_mcp_server.core.common.config.OPT_IN_TELEMETRY', False)
def test_user_agent_without_telemetry():
    """Test user agent when telemetry is disabled."""
    user_agent = get_user_agent_extra()
    assert 'cfg/ro#' not in user_agent
    assert 'cfg/consent#' not in user_agent
    assert 'cfg/scripts#' not in user_agent


@patch('awslabs.aws_api_mcp_server.core.common.config.get_context')
@patch('awslabs.aws_api_mcp_server.core.common.config.OPT_IN_TELEMETRY', False)
def test_user_agent_with_context(mock_get_context):
    """Test user agent when context is present."""
    # Create mock context with fastmcp name and client params
    mock_context = MagicMock()
    mock_context.fastmcp.name = 'test-fastmcp'
    mock_context.session.client_params.clientInfo.name = 'test-client'
    mock_context.session.client_params.clientInfo.version = '1.0.0'

    mock_get_context.return_value = mock_context

    user_agent = get_user_agent_extra()

    # Verify context information is included in user agent
    assert 'via/test-fastmcp' in user_agent
    assert 'MCPClient/test-client#1.0.0' in user_agent
    assert 'awslabs/mcp/AWS-API-MCP-server/' in user_agent


@patch('awslabs.aws_api_mcp_server.core.common.config.get_context')
@patch('awslabs.aws_api_mcp_server.core.common.config.OPT_IN_TELEMETRY', False)
def test_user_agent_with_context_no_client_params(mock_get_context):
    """Test user agent when context is present but client_params is None."""
    # Create mock context with fastmcp name but no client params
    mock_context = MagicMock()
    mock_context.fastmcp.name = 'test-fastmcp'
    mock_context.session.client_params = None

    mock_get_context.return_value = mock_context

    user_agent = get_user_agent_extra()

    # Verify fastmcp name is included but client info is not
    assert 'via/test-fastmcp' in user_agent
    assert 'MCPClient/' not in user_agent
    assert 'awslabs/mcp/AWS-API-MCP-server/' in user_agent


@patch('importlib.metadata.version')
def test_package_version_fallback_to_unknown(mock_version):
    """Test that PACKAGE_VERSION falls back to 'unknown' when package not found."""
    original_version = config_module.PACKAGE_VERSION

    mock_version.side_effect = PackageNotFoundError()
    importlib.reload(config_module)

    assert config_module.PACKAGE_VERSION == 'unknown'

    user_agent = config_module.get_user_agent_extra()
    assert 'awslabs/mcp/AWS-API-MCP-server/unknown' in user_agent

    # Restore original state
    config_module.PACKAGE_VERSION = original_version


@pytest.mark.parametrize(
    'env_value,expected_mode',
    [
        ('unrestricted', FileAccessMode.UNRESTRICTED),
        ('true', FileAccessMode.UNRESTRICTED),
        ('yes', FileAccessMode.UNRESTRICTED),
        ('1', FileAccessMode.UNRESTRICTED),
        ('Unrestricted', FileAccessMode.UNRESTRICTED),
        ('True', FileAccessMode.UNRESTRICTED),
        ('YES', FileAccessMode.UNRESTRICTED),
        ('TRUE', FileAccessMode.UNRESTRICTED),
    ],
)
def test_get_file_access_mode_unrestricted(monkeypatch, env_value, expected_mode):
    """Test that 'true', 'yes', '1' map to FileAccessMode.UNRESTRICTED (case-insensitive)."""
    monkeypatch.setenv('AWS_API_MCP_ALLOW_UNRESTRICTED_LOCAL_FILE_ACCESS', env_value)
    result = get_file_access_mode()
    assert result == expected_mode


@pytest.mark.parametrize(
    'env_value,expected_mode',
    [
        ('false', FileAccessMode.WORKDIR),
        ('no', FileAccessMode.WORKDIR),
        ('0', FileAccessMode.WORKDIR),
        ('workdir', FileAccessMode.WORKDIR),
        ('False', FileAccessMode.WORKDIR),
        ('NO', FileAccessMode.WORKDIR),
        ('WORKDIR', FileAccessMode.WORKDIR),
        ('Workdir', FileAccessMode.WORKDIR),
    ],
)
def test_get_file_access_mode_workdir(monkeypatch, env_value, expected_mode):
    """Test that 'false', 'no', '0', 'workdir' map to FileAccessMode.WORKDIR (case-insensitive)."""
    monkeypatch.setenv('AWS_API_MCP_ALLOW_UNRESTRICTED_LOCAL_FILE_ACCESS', env_value)
    result = get_file_access_mode()
    assert result == expected_mode


@pytest.mark.parametrize(
    'env_value,expected_mode',
    [
        ('no-access', FileAccessMode.NO_ACCESS),
        ('NO-ACCESS', FileAccessMode.NO_ACCESS),
        ('No-Access', FileAccessMode.NO_ACCESS),
    ],
)
def test_get_file_access_mode_no_access(monkeypatch, env_value, expected_mode):
    """Test that 'no-access' maps to FileAccessMode.NO_ACCESS (case-insensitive)."""
    monkeypatch.setenv('AWS_API_MCP_ALLOW_UNRESTRICTED_LOCAL_FILE_ACCESS', env_value)
    result = get_file_access_mode()
    assert result == expected_mode


def test_get_file_access_mode_default(monkeypatch):
    """Test that default value is FileAccessMode.WORKDIR when env var is not set."""
    monkeypatch.delenv('AWS_API_MCP_ALLOW_UNRESTRICTED_LOCAL_FILE_ACCESS', raising=False)
    result = get_file_access_mode()
    assert result == FileAccessMode.WORKDIR


@pytest.mark.parametrize(
    'env_value',
    [
        'invalid',
        'random',
        'unknown',
        '',
        'yes_access',
    ],
)
def test_get_file_access_mode_unknown_defaults_to_workdir(monkeypatch, env_value):
    """Test that unknown values default to FileAccessMode.WORKDIR."""
    monkeypatch.setenv('AWS_API_MCP_ALLOW_UNRESTRICTED_LOCAL_FILE_ACCESS', env_value)
    result = get_file_access_mode()
    assert result == FileAccessMode.WORKDIR



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pytest
from awslabs.aws_api_mcp_server.core.common.command_metadata import CommandMetadata
from awslabs.aws_api_mcp_server.core.common.config import WORKING_DIRECTORY, FileAccessMode
from awslabs.aws_api_mcp_server.core.common.errors import FilePathValidationError
from awslabs.aws_api_mcp_server.core.common.file_system_controls import (
    CUSTOM_FILE_PATH_ARGUMENTS,
    extract_file_paths_from_parameters,
    validate_file_path,
)
from awslabs.aws_api_mcp_server.core.parser.parser import ALLOWED_CUSTOM_OPERATIONS
from unittest.mock import patch


def test_safe_path_allowed():
    """Test that files within working directory are allowed."""
    safe_path = os.path.join(WORKING_DIRECTORY, 'safe_file.txt')
    result = validate_file_path(safe_path)
    assert result == safe_path


def test_unsafe_path_blocked():
    """Test that files outside working directory are blocked."""
    unsafe_path = '/tmp/unsafe_file.txt'
    with pytest.raises(FilePathValidationError):
        validate_file_path(unsafe_path)


@patch(
    'awslabs.aws_api_mcp_server.core.common.file_system_controls.FILE_ACCESS_MODE',
    FileAccessMode.UNRESTRICTED,
)
def test_unrestricted_access_allows_unsafe_path():
    """Test that FILE_ACCESS_MODE.UNRESTRICTED allows files outside working directory."""
    unsafe_path = '/tmp/unsafe_file.txt'
    result = validate_file_path(unsafe_path)
    assert result == unsafe_path


def test_all_custom_operations_have_file_path_arguments_entry():
    """Test that all custom commands must have explicitly listed file path arguments.

    This ensures that every custom operation in ALLOWED_CUSTOM_OPERATIONS has a corresponding
    entry in CUSTOM_FILE_PATH_ARGUMENTS, even if it's an empty list. This is important for
    maintaining awareness of which custom operations accept file paths and which don't.
    """
    missing_entries = []

    for service, operations in ALLOWED_CUSTOM_OPERATIONS.items():
        # Skip the wildcard service
        if service == '*':
            continue

        # Check if service exists in CUSTOM_FILE_PATH_ARGUMENTS
        if service not in CUSTOM_FILE_PATH_ARGUMENTS:
            missing_entries.append(
                f"Service '{service}' is missing from CUSTOM_FILE_PATH_ARGUMENTS"
            )
            continue

        # Check if all operations for this service have entries
        for operation in operations:
            if operation not in CUSTOM_FILE_PATH_ARGUMENTS[service]:
                missing_entries.append(
                    f"Operation '{operation}' for service '{service}' is missing from CUSTOM_FILE_PATH_ARGUMENTS"
                )

    assert not missing_entries, (
        'The following custom operations are missing from CUSTOM_FILE_PATH_ARGUMENTS:\n'
        + '\n'.join(missing_entries)
        + '\n\nAll custom operations must have an explicit entry in CUSTOM_FILE_PATH_ARGUMENTS, '
        + "even if it's an empty list (for operations that don't accept file paths)."
    )


def test_extract_file_paths_service_non_custom_operation():
    """Test extract_file_paths_from_parameters with non-custom operation."""
    command_metadata = CommandMetadata(
        service_sdk_name='lambda', service_full_sdk_name='AWS Lambda', operation_sdk_name='Invoke'
    )

    parameters = {
        'FunctionName': 'MyFunction',
        'Payload': '{"key": "value"}',
    }

    result = extract_file_paths_from_parameters(command_metadata, parameters)

    assert result == []


def test_extract_file_paths_param_value_is_list():
    """Test extract_file_paths_from_parameters when param_value is a list for file path arguments."""
    command_metadata = CommandMetadata(
        service_sdk_name='s3', service_full_sdk_name='s3', operation_sdk_name='sync'
    )

    parameters = {
        '--paths': [
            '/local/file1.txt',
            's3://bucket/key1',
        ],
        '--other-param': 'value',
    }

    result = extract_file_paths_from_parameters(command_metadata, parameters)

    # Should extract only the local file paths (remote paths are filtered out)
    expected = ['/local/file1.txt']
    assert result == expected


def test_extract_file_paths_parameter_none_value():
    """Test extract_file_paths_from_parameters with emr create-cluster and --configurations parameter set to None."""
    command_metadata = CommandMetadata(
        service_sdk_name='emr', service_full_sdk_name='emr', operation_sdk_name='create-cluster'
    )

    parameters = {
        '--configurations': None,
        '--name': 'MyCluster',
    }

    result = extract_file_paths_from_parameters(command_metadata, parameters)

    assert result == []



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
from .command_metadata import CommandMetadata
from .config import (
    FILE_ACCESS_MODE,
    FILE_ACCESS_MODE_KEY,
    WORKING_DIRECTORY,
    FileAccessMode,
)
from .errors import FilePathValidationError, LocalFileAccessDisabledError
from awscli.arguments import CLIArgument
from awscli.paramfile import get_file
from pathlib import Path
from typing import Any, Dict, List


# Regex pattern for file:// or fileb:// prefixes used in blob arguments
FILE_BLOB_PREFIX_PATTERN = r'^fileb?://'

# Custom operation arguments that accept file paths
# This includes empty entries for all allowed customizations that don't accept file paths
# to make sure that no opearation is overlooked
CUSTOM_FILE_PATH_ARGUMENTS = {
    's3': {
        'ls': [],
        'website': [],
        'sync': ['--paths'],
        'cp': ['--paths'],
        'mv': ['--paths'],
        'rm': [],
        'mb': [],
        'rb': [],
        'presign': [],
    },
    'cloudformation': {
        'package': ['--template-file', '--output-template-file'],
        'deploy': ['--template-file'],
    },
    'cloudfront': {'sign': ['--private-key']},
    'cloudtrail': {
        'validate-logs': [],
    },
    'codeartifact': {'login': []},
    'codecommit': {'credential-helper': []},
    'datapipeline': {'list-runs': [], 'create-default-roles': []},
    'dlm': {'create-default-role': []},
    'ecr': {'get-login': [], 'get-login-password': []},
    'ecr-public': {'get-login-password': []},
    'ecs': {'deploy': ['--task-definition', '--codedeploy-appspec']},
    'eks': {'update-kubeconfig': ['--kubeconfig'], 'get-token': []},
    'emr': {
        'add-instance-groups': [],
        'describe-cluster': [],
        'terminate-clusters': [],
        'modify-cluster-attributes': [],
        'install-applications': [],
        'create-cluster': [],
        'add-steps': [],
        'restore-from-hbase-backup': [],
        'create-hbase-backup': [],
        'schedule-hbase-backup': [],
        'disable-hbase-backups': [],
        'create-default-roles': [],
    },
    'emr-containers': {'update-role-trust-policy': []},
    'gamelift': {'upload-build': ['--build-root'], 'get-game-session-log': ['--save-as']},
    'rds': {'generate-db-auth-token': []},
    'servicecatalog': {'generate': ['--file-path']},
    'deploy': {'push': ['--source'], 'register': [], 'deregister': []},
    'configservice': {'subscribe': [], 'get-status': []},
}

# Custom operation arguments that accept file paths with the file:// or fileb:// prefixes
CUSTOM_BLOB_ARGUMENTS = {
    'emr': {
        'create-cluster': [
            '--configurations',
            '--bootstrap-actions',
            '--ec2-attributes',
            '--instance-groups',
            '--instance-fleets',
            '--kerberos-attributes',
            '--managed-scaling-policy',
            '--placement-group-configs',
            '--auto-termination-policy',
            '--additional-info',
            '--emrfs',
        ],
        'add-steps': [
            '--steps',
        ],
    },
}


def is_streaming_blob_argument(cli_argument: CLIArgument) -> bool:
    """Streaming blob arguments accept only file paths."""
    argument_model = cli_argument.argument_model
    return argument_model.type_name == 'blob' and argument_model.serialization.get('streaming')


def get_file_validated(prefix, path, mode):
    """Validate that a URI path (i.e. file://<path>) is within the allowed working directory."""
    file_path = os.path.expandvars(os.path.expanduser(path[len(prefix) :]))
    validate_file_path(file_path)

    return get_file(prefix, path, mode)


def validate_file_path(file_path: str) -> str:
    """Validate that a file path is within the allowed working directory.

    Args:
        file_path: The file path to validate

    Returns:
        The validated absolute path

    Raises:
        LocalFileAccessDisabledError: If local file access is disabled
        FilePathValidationError: If the path is outside the working directory and unrestricted access is not allowed
    """
    if FILE_ACCESS_MODE == FileAccessMode.NO_ACCESS:
        # Reject local file paths
        raise LocalFileAccessDisabledError(file_path)

    if FILE_ACCESS_MODE == FileAccessMode.UNRESTRICTED:
        return file_path

    # Convert to absolute path
    absolute_path = os.path.abspath(file_path)
    working_directory = os.path.abspath(WORKING_DIRECTORY)

    # Check if the path is within the working directory
    try:
        Path(absolute_path).resolve().relative_to(Path(working_directory).resolve())
    except ValueError:
        reason = (
            f"is outside the allowed working directory '{WORKING_DIRECTORY}'. "
            f'Set {FILE_ACCESS_MODE_KEY}=unrestricted to allow unrestricted file access.'
        )
        raise FilePathValidationError(file_path, reason)

    return absolute_path


def extract_file_paths_from_parameters(
    command_metadata: CommandMetadata, parameters: Dict[str, Any]
) -> List[str]:
    """Extract all potential file paths from custom command parameters.

    NOTE: this function only handles AWS CLI customizations (e.g. aws s3 cp, aws cloudformation package).
    Regular service operations are handled separately.

    This function extracts file paths from both regular file path arguments and blob arguments
    (with file:// or fileb:// prefixes). For blob arguments, the prefixes are removed.

    Args:
        command_metadata: Metadata about the command being executed
        parameters: Dictionary of command parameters

    Returns:
        List of file paths (with file:// and fileb:// prefixes removed)
    """
    file_paths = []
    service = command_metadata.service_sdk_name
    operation = command_metadata.operation_sdk_name

    # Get file path arguments for this service/operation
    file_path_args = set()
    if service in CUSTOM_FILE_PATH_ARGUMENTS and operation in CUSTOM_FILE_PATH_ARGUMENTS[service]:
        file_path_args = set(CUSTOM_FILE_PATH_ARGUMENTS[service][operation])

    # Get blob arguments for this service/operation
    blob_args = set()
    if service in CUSTOM_BLOB_ARGUMENTS and operation in CUSTOM_BLOB_ARGUMENTS[service]:
        blob_args = set(CUSTOM_BLOB_ARGUMENTS[service][operation])

    # Extract file paths from parameters
    for param_name, param_value in parameters.items():
        # Check if this is a file path argument
        if param_name in file_path_args:
            if isinstance(param_value, str) and not _is_remote_path(param_value):
                file_paths.append(param_value)
            elif isinstance(param_value, list):
                file_paths.extend(
                    [
                        item
                        for item in param_value
                        if isinstance(item, str) and not _is_remote_path(item)
                    ]
                )

        # Check if this is a blob argument (may have file:// or fileb:// prefix)
        elif param_name in blob_args:
            if isinstance(param_value, str):
                # Remove file:// or fileb:// prefix if present
                if re.match(FILE_BLOB_PREFIX_PATTERN, param_value):
                    cleaned_path = re.sub(FILE_BLOB_PREFIX_PATTERN, '', param_value)
                    file_paths.append(cleaned_path)
            elif isinstance(param_value, list):
                file_paths.extend(
                    [
                        re.sub(FILE_BLOB_PREFIX_PATTERN, '', item)
                        for item in param_value
                        if isinstance(item, str) and re.match(FILE_BLOB_PREFIX_PATTERN, item)
                    ]
                )

    return file_paths


def _is_remote_path(path: str) -> bool:
    """Check if path is remote (S3, HTTP, etc.)."""
    return path.startswith(('s3://', 'http://', 'https://', 'ftp://', 'arn:'))



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import botocore.serialize
import ipaddress
import jmespath
import os
import re
from ..aws.regions import GLOBAL_SERVICE_REGIONS
from ..aws.services import (
    get_awscli_driver,
    get_operation_filters,
)
from ..common.command import IRCommand, OutputFile
from ..common.command_metadata import CommandMetadata
from ..common.config import AWS_API_MCP_PROFILE_NAME, FILE_ACCESS_MODE, FileAccessMode, get_region
from ..common.errors import (
    AwsApiMcpError,
    ClientSideFilterError,
    CommandValidationError,
    DeniedGlobalArgumentsError,
    ExpectedArgumentError,
    FileParameterError,
    FilePathValidationError,
    InvalidChoiceForParameterError,
    InvalidParametersReceivedError,
    InvalidServiceError,
    InvalidServiceOperationError,
    InvalidTypeForParameterError,
    LocalFileAccessDisabledError,
    MalformedFilterError,
    MissingOperationError,
    MissingRequiredParametersError,
    MisspelledParametersError,
    OperationNotAllowedError,
    ParameterSchemaValidationError,
    ParameterValidationErrorRecord,
    RequestSerializationError,
    ServiceNotAllowedError,
    ShortHandParserError,
    UnknownArgumentsError,
    UnknownFiltersError,
    UnsupportedFilterError,
)
from ..common.file_system_controls import extract_file_paths_from_parameters, validate_file_path
from ..common.helpers import expand_user_home_directory, is_help_operation
from .custom_validators.botocore_param_validator import BotoCoreParamValidator
from .custom_validators.ec2_validator import validate_ec2_parameter_values
from .custom_validators.ssm_validator import perform_ssm_validations
from .lexer import split_cli_command
from argparse import Namespace
from awscli.argparser import ArgTableArgParser, CommandAction, MainArgParser
from awscli.argprocess import ParamError
from awscli.arguments import BaseCLIArgument, CLIArgument
from awscli.clidriver import ServiceCommand
from botocore.exceptions import ParamValidationError, UndefinedModelAttributeError
from botocore.model import OperationModel, ServiceModel
from collections.abc import Generator
from difflib import SequenceMatcher
from jmespath.exceptions import ParseError
from pathlib import Path
from typing import Any, NamedTuple, cast
from urllib.parse import urlparse


ARN_PATTERN = re.compile(
    r'^(arn:(?:aws|aws-cn|aws-iso|aws-iso-b|aws-iso-e|aws-iso-f|aws-us-gov):[\w\d-]+:([\w\d-]*):\d{0,12}:[\w\d-]*\/?[\w\d-]*)(\/.*)?.*$'
)

# These are subcommands for `aws` which are not actual services.
# They are not ServiceCommand instances. The other example of a non-ServiceCommand
# is the fake "s3" service, which is handled properly.
DENIED_CUSTOM_SERVICES = frozenset({'configure', 'history'})

# These are the custom operations for `aws` services in CLI which are known
# to not do any subprocess calls and are therefore allowed.
ALLOWED_CUSTOM_OPERATIONS = {
    # blanket allow these custom operation regardless of service
    '*': [],
    's3': ['ls', 'website', 'sync', 'cp', 'mv', 'rm', 'mb', 'rb', 'presign'],
    'cloudformation': ['package', 'deploy'],
    'cloudfront': ['sign'],
    'cloudtrail': ['validate-logs'],
    'codeartifact': ['login'],
    'codecommit': ['credential-helper'],
    'datapipeline': ['list-runs', 'create-default-roles'],
    'dlm': ['create-default-role'],
    'ecr': ['get-login', 'get-login-password'],
    'ecr-public': ['get-login-password'],
    'ecs': ['deploy'],
    'eks': ['update-kubeconfig', 'get-token'],
    'emr': [
        'add-instance-groups',
        'describe-cluster',
        'terminate-clusters',
        'modify-cluster-attributes',
        'install-applications',
        'create-cluster',
        'add-steps',
        'restore-from-hbase-backup',
        'create-hbase-backup',
        'schedule-hbase-backup',
        'disable-hbase-backups',
        'create-default-roles',
    ],
    'emr-containers': ['update-role-trust-policy'],
    'gamelift': ['upload-build', 'get-game-session-log'],
    'rds': ['generate-db-auth-token'],
    'servicecatalog': ['generate'],
    'deploy': ['push', 'register', 'deregister'],
    'configservice': ['subscribe', 'get-status'],
}

# These are the custom operations allowed when local file access is disabled.
# This is a subset of ALLOWED_CUSTOM_OPERATIONS that excludes operations requiring local file access.
ALLOWED_CUSTOM_OPERATIONS_WHEN_FILE_ACCESS_DISABLED = {
    # blanket allow these custom operation regardless of service
    '*': [],
    's3': ['ls', 'website', 'sync', 'cp', 'mv', 'rm', 'mb', 'rb', 'presign'],
    'cloudtrail': ['validate-logs'],
    'codecommit': ['credential-helper'],
    'datapipeline': ['list-runs', 'create-default-roles'],
    'dlm': ['create-default-role'],
    'ecr': ['get-login', 'get-login-password'],
    'ecr-public': ['get-login-password'],
    'eks': ['get-token'],
    'emr': [
        'add-instance-groups',
        'create-cluster',
        'describe-cluster',
        'terminate-clusters',
        'modify-cluster-attributes',
        'install-applications',
        'add-steps',
        'restore-from-hbase-backup',
        'create-hbase-backup',
        'schedule-hbase-backup',
        'disable-hbase-backups',
        'create-default-roles',
    ],
    'emr-containers': ['update-role-trust-policy'],
    'rds': ['generate-db-auth-token'],
    'deploy': ['deregister'],
    'configservice': ['subscribe', 'get-status'],
}

_excluded_optional_params = frozenset(
    {
        '--cli-input-json',
        '--generate-cli-skeleton',
        '--dry-run',
        '--no-dry-run',
    }
)

NARGS_ONE_ARGUMENT = None
NARGS_OPTIONAL = '?'
NARGS_ONE_OR_MORE = '+'

# Map nargs (number of time arguments can appear from argparse point of view)
# to the corresponding error. These are implicitly defined in argparse.
_nargs_errors = {
    NARGS_ONE_ARGUMENT: 'expected one argument',
    NARGS_OPTIONAL: 'expected at most one argument',
    NARGS_ONE_OR_MORE: 'expected at least one argument',
}

ALLOWED_FILTER_KEYS_SUBSETS = {
    frozenset({'Name', 'Values'}): 'Name',
    frozenset({'Key', 'Values'}): 'Key',
    frozenset({'key', 'value'}): 'key',
}


class ParsedOperationArgs(NamedTuple):
    """Named tuple to store parsed operation arguments."""

    operation_args: Namespace
    supported_args: list[str]
    given_args: list[str]
    missing_parameters: list[str]
    unknown_parameters: list[str]
    unknown_args: list[str]


def _on_error_in_argparse(message: str):
    raise AwsApiMcpError(message)


class ArgTableParser(ArgTableArgParser):
    """Parser for argument tables, supporting AWS CLI command metadata."""

    def parse_operation_args(self, command_metadata: CommandMetadata, args: list[str]):
        """Parse known arguments using the provided command metadata and argument list."""
        self.command_metadata = command_metadata
        operation_args, unknown_args = super().parse_known_args(args)

        supported_args = [
            action.option_strings[0] for action in self._actions if action.option_strings
        ]

        missing_parameters = list(self._identify_missing_parameters(operation_args))

        return ParsedOperationArgs(
            operation_args=operation_args,
            supported_args=supported_args,
            given_args=args,
            missing_parameters=missing_parameters,
            unknown_parameters=[
                param
                for param in unknown_args
                if param.startswith('-')
                and param not in supported_args
                and not any(arg.startswith(param) for arg in supported_args if self.allow_abbrev)
            ],
            unknown_args=[param for param in unknown_args if not param.startswith('-')],
        )

    def _check_if_misspelled(self, service, operation, supported_args, unknown_args):
        for unknown_arg in unknown_args:
            if unknown_arg.startswith('--'):
                for supported_arg in supported_args:
                    similarity = SequenceMatcher(None, supported_arg, unknown_arg).ratio()
                    if similarity >= 0.8:
                        raise MisspelledParametersError(
                            service=service,
                            operation=operation,
                            unknown_parameter=unknown_arg,
                            existing_parameter=supported_arg,
                        )

    def error(self, message):  # type: ignore[override]
        """Handle errors during argument parsing."""
        # Skip throwing errors to collate all fields that are missing/not recognized
        pass

    def _identify_missing_parameters(self, operation_args: Namespace) -> Generator[str]:
        # Check for required named arguments (those with option_strings)
        required_named_args = {
            action.option_strings[0]
            for action in self._actions
            if action.option_strings and action.required
        }

        # Check for required positional arguments (those without option_strings but with nargs)
        required_positional_args = {
            action.dest
            for action in self._actions
            if not action.option_strings
            and action.nargs
            and action.nargs != '?'
            and action.nargs != '*'
        }

        for name, value in vars(operation_args).items():
            if value is None:
                # Check if it's a required named argument
                cli_param = f'--{name.replace("_", "-")}'
                if cli_param in required_named_args:
                    yield cli_param
                # Check if it's a required positional argument
                elif name in required_positional_args:
                    yield name

    def _get_value(self, action, arg_string):
        try:
            return super()._get_value(action, arg_string)
        except argparse.ArgumentError as exc:
            raise InvalidTypeForParameterError(action.option_strings[0], action.type) from exc  # type: ignore

    def _match_argument(self, action, arg_strings_pattern):
        try:
            return super()._match_argument(action, arg_strings_pattern)
        except argparse.ArgumentError as exc:
            msg: str = _fetch_error_from_number_of_args(action.nargs)  # type: ignore
            raise ExpectedArgumentError(
                action.option_strings[0], msg, self.command_metadata
            ) from exc


def _fetch_error_from_number_of_args(nargs: str) -> str:
    return cast(str, _nargs_errors.get(nargs))


class GlobalArgParser(MainArgParser):
    """Parser for global AWS CLI arguments."""

    def _check_value(self, action, value):
        """Check if the value is valid for the given action."""
        if action.choices is not None and value not in action.choices:
            if action.dest == 'command':
                # This service does not exist. The command table contains service aliases
                # as well (e.g. `s3` is not an actual "service" in the underlying model, `s3api` is.
                raise InvalidServiceError(value)
            raise InvalidChoiceForParameterError(action.dest, value)
        return super()._check_value(action, value)

    # Overwrite _build's parent method as it automatically injects a `version` action in the
    # parser. Version actions print the current version and then exit the program, which is
    # not what we want.
    def _build(self, command_table, version_string, argument_table):  # noqa: ARG002
        for argument_name in argument_table:
            argument = argument_table[argument_name]
            argument.add_to_parser(self)
        self.add_argument('--version')
        self.add_argument('command', action=CommandAction, command_table=command_table)

    @staticmethod
    def get_parser():
        """Return a new instance of GlobalArgParser."""
        return GlobalArgParser(
            command_table,
            session.user_agent(),
            cli_data.get('description', None),
            driver._get_argument_table(),
            prog='aws',
        )

    def error(self, message):  # type: ignore[override]
        """Handle errors in global argument parsing."""
        _on_error_in_argparse(message)


def is_custom_operation(service, operation):
    """Returns true if the service operation is cli customization."""
    service_command = command_table.get(service, None)
    if not service_command:
        raise InvalidServiceError(service)

    if isinstance(service_command, ServiceCommand):
        # valid service, unlike s3
        service_command_table = service_command._get_command_table()
        operation_command = service_command_table.get(operation)

        # valid service can have custom operations.
        # custom operations don't have _operation_model
        if hasattr(operation_command, '_operation_model'):
            return False

    return True


def is_denied_custom_service(service):
    """Returns true if the service is a cli customization that is explicitely denied."""
    return service in DENIED_CUSTOM_SERVICES


def is_denied_custom_operation(service, operation):
    """Check if a service operation is custom and denied."""
    if not is_custom_operation(service, operation):
        return False

    # Choose the appropriate allowlist based on file access settings
    allowed_operations = (
        ALLOWED_CUSTOM_OPERATIONS_WHEN_FILE_ACCESS_DISABLED
        if FILE_ACCESS_MODE == FileAccessMode.NO_ACCESS
        else ALLOWED_CUSTOM_OPERATIONS
    )

    if operation in allowed_operations['*']:
        return False

    return not (service in allowed_operations and operation in allowed_operations[service])


driver = get_awscli_driver()
session = driver.session
command_table = driver._get_command_table()
cli_data = driver._get_cli_data()
parser = GlobalArgParser.get_parser()
driver._add_aliases(command_table, parser)


def parse(cli_command: str, default_region_override: str | None = None) -> IRCommand:
    """Parse a CLI command string into an IRCommand object."""
    tokens = split_cli_command(cli_command)
    # Strip `aws` and expand paths beginning with ~
    tokens = expand_user_home_directory(tokens[1:])
    service_namespace, args = parser.parse_known_args(tokens)
    service_command = command_table[service_namespace.command]

    if service_command.name in DENIED_CUSTOM_SERVICES:
        raise ServiceNotAllowedError(service_command.name)

    if isinstance(service_command, ServiceCommand):
        return _handle_service_command(
            service_command, service_namespace, args, default_region_override
        )

    return _handle_awscli_customization(
        service_namespace, args, tokens[0], default_region_override
    )


def _handle_service_command(
    service_command: ServiceCommand,
    global_args: argparse.Namespace,
    remaining: list[str],
    default_region_override: str | None = None,
):
    if not remaining:
        raise MissingOperationError()

    service = service_command.name
    command_table = service_command._get_command_table()

    operation = remaining[0]
    operation_command = command_table.get(operation)
    if not operation_command:
        # This command is not supported for this service
        raise InvalidServiceOperationError(service, operation)
    if not hasattr(operation_command, '_operation_model'):
        return _handle_awscli_customization(global_args, remaining, service_command.name)
    command_metadata = CommandMetadata(
        service_sdk_name=service_command.service_model.service_name,
        service_full_sdk_name=_service_full_name(service_command.service_model),
        operation_sdk_name=operation_command._operation_model.name,
        has_streaming_output=operation_command._operation_model.has_streaming_output,
    )
    _validate_global_args(service, global_args)
    region = getattr(global_args, 'region', None)

    service_parser = service_command._create_parser()
    service_args, service_remaining = service_parser.parse_known_args(remaining)
    operation_parser = ArgTableParser(operation_command.arg_table)
    parsed_args = operation_parser.parse_operation_args(command_metadata, service_remaining)
    _handle_invalid_parameters(command_metadata, service, operation, parsed_args)

    try:
        parameters = operation_command._build_call_parameters(
            parsed_args.operation_args, operation_command.arg_table
        )
    except ParamError as exc:
        raise ShortHandParserError(exc.cli_name, exc.message) from exc
    except CommandValidationError:
        raise
    except Exception as exc:
        raise CommandValidationError(exc) from exc

    _validate_filters(
        service_command.service_model.service_name,
        operation,
        operation_command._operation_model,
        parameters,
    )

    _validate_parameters(
        parameters, operation_command.arg_table, operation_command._operation_model
    )

    arn_region = _fetch_region_from_arn(parameters)
    global_args.region = region or arn_region
    if (
        command_metadata.service_sdk_name in GLOBAL_SERVICE_REGIONS
        and global_args.region != GLOBAL_SERVICE_REGIONS[command_metadata.service_sdk_name]
    ):
        global_args.region = GLOBAL_SERVICE_REGIONS[command_metadata.service_sdk_name]

    _validate_outfile(command_metadata, parsed_args)

    _validate_request_serialization(
        operation,
        service_command.service_model,
        operation_command._operation_model,
        parameters,
    )

    _run_custom_validations(
        service_command.service_model.service_name,
        operation,
        parameters,
    )

    return _construct_command(
        command_metadata=command_metadata,
        global_args=global_args,
        parameters=parameters,
        parsed_args=parsed_args,
        operation_model=operation_command._operation_model,
        default_region_override=default_region_override,
    )


def _handle_awscli_customization(
    global_args: argparse.Namespace,
    remaining: list[str],
    service: str,
    default_region_override: str | None = None,
) -> IRCommand:
    """This function handles awscli customizations (like aws s3 ls, aws s3 cp, aws s3 mv)."""
    if not remaining:
        raise MissingOperationError()

    operation = remaining[0]

    command_table = driver._get_command_table()
    service_command = command_table.get(service)

    if service_command is None:
        raise InvalidServiceError(service)

    # For custom commands, we need to check if the operation exists in the service's command table
    if hasattr(service_command, '_get_command_table'):
        service_command_table = service_command._get_command_table()
        operation_command = service_command_table.get(operation)
    elif hasattr(service_command, 'subcommand_table'):
        # Handle S3-like services that use subcommand_table
        service_command_table = service_command.subcommand_table
        operation_command = service_command_table.get(operation)
    else:
        raise InvalidServiceOperationError(service, operation)

    if not operation_command:
        raise InvalidServiceOperationError(service, operation)

    if is_denied_custom_operation(service, operation):
        raise OperationNotAllowedError(service, operation)

    if not hasattr(operation_command, '_operation_model'):
        return _validate_customization_arguments(
            operation_command, global_args, remaining, service, operation, default_region_override
        )

    raise InvalidServiceOperationError(service, operation)


def contains_subcommand(operation_command, remaining: list[str]) -> bool:
    """Check if the operation command has subcommands and the remaining args contain a subcommand."""
    return (
        hasattr(operation_command, 'subcommand_table')
        and operation_command.subcommand_table
        and len(remaining) > 1
        and not remaining[1].startswith('--')
    )


def _parse_customization_parameters(
    operation_command,
    command_metadata: CommandMetadata,
    operation_args: list[str],
    service: str,
    operation: str,
) -> dict[str, Any]:
    """Parse parameters for a custom command using its argument table."""
    if not hasattr(operation_command, 'arg_table'):
        raise InvalidServiceOperationError(service, operation)

    operation_parser = ArgTableParser(operation_command.arg_table)
    parsed_args = operation_parser.parse_operation_args(command_metadata, operation_args)

    _handle_invalid_parameters(command_metadata, service, operation, parsed_args)

    parameters = {
        f'--{key.replace("_", "-")}': value
        for key, value in vars(parsed_args.operation_args).items()
        if value is not None
    }

    return parameters


def _validate_customization_arguments(
    operation_command,
    global_args: argparse.Namespace,
    remaining: list[str],
    service: str,
    operation: str,
    default_region_override: str | None = None,
) -> IRCommand:
    """Validate arguments for awscli customizations using their argument table."""
    _validate_global_args(service, global_args)
    global_args.region = getattr(global_args, 'region', None)

    if contains_subcommand(operation_command, remaining):
        subcommand_name = remaining[1]
        subcommand = operation_command.subcommand_table.get(subcommand_name)

        if not subcommand:
            raise InvalidServiceOperationError(service, f'{operation} {subcommand_name}')

        # Update the operation name to include the subcommand
        full_operation = f'{operation} {subcommand_name}'
        command_metadata = CommandMetadata(
            service_sdk_name=service,
            service_full_sdk_name=None,
            operation_sdk_name=full_operation,
        )

        # Parse the remaining arguments (skip the operation and subcommand names)
        operation_args = remaining[2:] if len(remaining) > 2 else []
        parameters = _parse_customization_parameters(
            subcommand, command_metadata, operation_args, service, full_operation
        )

        # Validate file paths for custom commands with subcommands
        _validate_customization_file_paths(command_metadata, service, full_operation, parameters)

        return _construct_command(
            command_metadata=command_metadata,
            global_args=global_args,
            parameters=parameters,
            is_awscli_customization=True,
            default_region_override=default_region_override,
        )
    else:
        # This is a regular custom command without subcommands (or invalid subcommand)
        # Parse the remaining arguments (skip the operation name)
        command_metadata = CommandMetadata(
            service_sdk_name=service,
            service_full_sdk_name=None,
            operation_sdk_name=operation,
        )

        operation_args = remaining[1:] if len(remaining) > 1 else []
        parameters = _parse_customization_parameters(
            operation_command, command_metadata, operation_args, service, operation
        )

        # Run custom validations for S3 customizations
        if service == 's3':
            _validate_s3_file_paths(service, operation, parameters)
        else:
            # Validate file paths for other custom commands
            _validate_customization_file_paths(command_metadata, service, operation, parameters)

        return _construct_command(
            command_metadata=command_metadata,
            global_args=global_args,
            parameters=parameters,
            is_awscli_customization=True,
            default_region_override=default_region_override,
        )


def _handle_invalid_parameters(
    command_metadata: CommandMetadata,
    service: str,
    operation: str,
    parsed_args: ParsedOperationArgs,
):
    # Exclude a set of parameters that are not supported
    supported_parameters_with_exclusions = (
        set(parsed_args.supported_args) - _excluded_optional_params
    )

    if parsed_args.unknown_parameters:
        raise InvalidParametersReceivedError(
            service=service,
            operation=operation,
            invalid_parameters=sorted(parsed_args.unknown_parameters),
            correct_parameters=sorted(supported_parameters_with_exclusions),
        )
    if parsed_args.missing_parameters:
        raise MissingRequiredParametersError(
            service=service,
            operation=operation,
            parameters=parsed_args.missing_parameters,
            command_metadata=command_metadata,
        )
    if parsed_args.unknown_args:
        raise UnknownArgumentsError(
            service=service,
            operation=operation,
            unknown_args=parsed_args.unknown_args,
        )


def _validate_global_args(service: str, global_args: argparse.Namespace):
    denied_args = []
    if global_args.debug:
        denied_args.append('--debug')
    if not global_args.verify_ssl:
        denied_args.append('--no-verify-ssl')
    if not global_args.sign_request:
        denied_args.append('--no-sign-request')
    if denied_args:
        raise DeniedGlobalArgumentsError(service, sorted(denied_args))


def _validate_parameters(
    parameters: dict[str, Any],
    arg_table: dict[str, BaseCLIArgument],
    operation_model: OperationModel,
) -> None:
    validator = BotoCoreParamValidator()

    serialized_to_cli = {
        arg._serialized_name: arg.cli_name
        for arg in arg_table.values()
        if isinstance(arg, CLIArgument)
        and hasattr(arg, '_serialized_name')
        and hasattr(arg, 'cli_name')
    }

    errors = []

    input_shape = operation_model.input_shape
    boto3_members = getattr(input_shape, 'members', {})

    for key, value in parameters.items():
        boto3_shape = boto3_members.get(key)
        if boto3_shape is not None:
            report = validator.validate(value, boto3_shape)
            if report.has_errors():
                cli_name = serialized_to_cli.get(key, key)
                errors.append(ParameterValidationErrorRecord(cli_name, report.generate_report()))
    if errors:
        raise ParameterSchemaValidationError(errors)


def _validate_filters(
    service: str, operation: str, operation_model: OperationModel, parameters: dict[str, Any]
):
    if 'Filters' not in parameters:
        return

    filters = parameters['Filters']
    known_filters = get_operation_filters(operation_model)

    filter_name_key = None
    for allowed_keys_subset, name_key in ALLOWED_FILTER_KEYS_SUBSETS.items():
        if allowed_keys_subset.issubset(known_filters.filter_keys):
            filter_name_key = name_key

    if filter_name_key is None:
        raise UnsupportedFilterError(service, operation, known_filters.filter_keys)

    unknown_filters = []
    for filter_element in filters:
        filter_element_key_set = filter_element.keys()
        if filter_element_key_set != known_filters.filter_keys:
            raise MalformedFilterError(
                service, operation, filter_element_key_set, known_filters.filter_keys
            )

        filter_name = filter_element.get(filter_name_key)
        if not known_filters.allows_filter(filter_name):
            unknown_filters.append(filter_name)

    if unknown_filters:
        raise UnknownFiltersError(service, sorted(unknown_filters))


def _run_custom_validations(service: str, operation: str, parameters: dict[str, Any]):
    if service == 'ssm':
        perform_ssm_validations(operation, parameters)
    if service == 'ec2':
        validate_ec2_parameter_values(parameters)


def _validate_request_serialization(
    operation: str,
    service_model: ServiceModel,
    operation_model: OperationModel,
    parameters: dict[str, Any],
):
    validated_parameters = parameters.copy()
    validated_parameters.pop('PaginationConfig', None)

    # Parameter validation has been done, just serialize
    serializer = botocore.serialize.create_serializer(
        service_model.metadata['protocol'], include_validation=False
    )
    try:
        serializer.serialize_to_request(validated_parameters, operation_model)
    except ParamValidationError as err:
        raise RequestSerializationError(
            str(service_model.service_name), operation, str(err)
        ) from err


def _validate_s3_file_paths(service: str, operation: str, parameters: dict[str, Any]):
    if operation not in ('cp', 'sync', 'mv'):
        return

    paths = parameters.get('--paths')
    if not paths or not isinstance(paths, list) or len(paths) < 2:
        return

    source_path, dest_path = paths
    _validate_s3_file_path(source_path, service, operation)
    _validate_s3_file_path(dest_path, service, operation)


def _validate_s3_file_path(file_path: str, service: str, operation: str):
    if file_path == '-':
        raise FileParameterError(
            service=service,
            operation=operation,
            file_path=file_path,
            reason="streaming file ('-') is not allowed",
        )

    if not file_path.startswith('s3://'):
        _validate_file_path(file_path, service, operation)


def _validate_customization_file_paths(
    command_metadata: CommandMetadata,
    service: str,
    operation: str,
    parameters: dict[str, Any],
):
    """Validate file paths in custom command parameters.

    This function extracts file paths from custom command parameters (both regular
    file path arguments and blob arguments with file:// or fileb:// prefixes) and
    validates each one through _validate_file_path.

    Args:
        command_metadata: Metadata about the command being executed
        service: The AWS service name
        operation: The operation name
        parameters: Dictionary of command parameters

    Raises:
        FileParameterError: If any file path validation fails
    """
    # Extract all file paths from parameters (with prefixes removed)
    file_paths = extract_file_paths_from_parameters(command_metadata, parameters)

    # Validate each file path
    for file_path in file_paths:
        _validate_file_path(file_path, service, operation)


def _validate_outfile(
    command_metadata: CommandMetadata,
    parsed_args: ParsedOperationArgs | None,
):
    """Validate streaming outfile argument."""
    # Validate positional outfile argument for streaming operations
    if command_metadata.has_streaming_output and parsed_args:
        output_file_path = parsed_args.operation_args.outfile
        if output_file_path != '-':
            _validate_file_path(
                output_file_path,
                service=command_metadata.service_sdk_name,
                operation=command_metadata.operation_sdk_name,
            )


def _validate_file_path(file_path: str, service: str, operation: str):
    if not os.path.isabs(Path(file_path)):
        raise FileParameterError(
            service=service,
            operation=operation,
            file_path=file_path,
            reason='should be an absolute path',
        )

    try:
        validate_file_path(file_path)
    except (FilePathValidationError, LocalFileAccessDisabledError) as e:
        raise FileParameterError(
            service=service,
            operation=operation,
            file_path=file_path,
            reason=e._reason,
        )


def _validate_endpoint(endpoint: str | None):
    if not endpoint:
        return

    try:
        url = urlparse(endpoint if '://' in endpoint else f'http://{endpoint}')
        url.port  # will throw an exception if the port is not a number
    except Exception as e:
        raise ValueError(f'Invalid endpoint or port: {endpoint}') from e

    hostname = url.hostname
    if not hostname:
        raise ValueError(f'Could not find hostname {endpoint}')

    if hostname == 'localhost':
        hostname = '127.0.0.1'

    try:
        ip_obj = ipaddress.ip_address(hostname)
        if not ip_obj.is_loopback:
            raise ValueError(f'Local endpoint was not a loopback address: {hostname}')
    except ValueError as e:
        raise ValueError(f'Could not resolve endpoint: {e}')


def _fetch_region_from_arn(parameters: dict[str, Any]) -> str | None:
    for param_value in parameters.values():
        if isinstance(param_value, str):
            m = ARN_PATTERN.match(param_value)
            if m and m.groups()[1]:
                return m.groups()[1]
    return None


def _construct_command(
    command_metadata: CommandMetadata,
    global_args: argparse.Namespace,
    parameters: dict[str, Any],
    is_awscli_customization: bool = False,
    parsed_args: ParsedOperationArgs | None = None,
    operation_model: OperationModel | None = None,
    default_region_override: str | None = None,
) -> IRCommand:
    _validate_outfile(command_metadata, parsed_args)
    endpoint_url = getattr(global_args, 'endpoint_url', None)
    _validate_endpoint(endpoint_url)

    explicitly_passed_arguments = list(parameters.values()) + (
        parsed_args.given_args if parsed_args else []
    )

    profile = getattr(global_args, 'profile', None)
    region = (
        getattr(global_args, 'region', None)
        or _fetch_region_from_arn(parameters)
        or default_region_override
        or get_region(profile or AWS_API_MCP_PROFILE_NAME)
    )

    client_side_query = getattr(global_args, 'query', None)
    client_side_filter = None

    if client_side_query is not None:
        try:
            client_side_filter = jmespath.compile(client_side_query)
        except ParseError as error:
            raise ClientSideFilterError(
                service=command_metadata.service_sdk_name,
                operation=command_metadata.operation_sdk_name,
                client_side_query=client_side_query,
                msg=str(error),
            )

    output_file = (
        OutputFile.from_operation(parsed_args.operation_args.outfile, operation_model)
        if command_metadata.has_streaming_output and parsed_args and operation_model
        else None
    )

    return IRCommand(
        command_metadata=command_metadata,
        parameters=parameters,
        region=region,
        profile=profile,
        client_side_filter=client_side_filter,
        is_awscli_customization=is_awscli_customization,
        is_help_operation=is_help_operation(explicitly_passed_arguments),
        output_file=output_file,
        endpoint_url=global_args.endpoint_url,
    )


def _service_full_name(service_model: ServiceModel) -> str | None:
    try:
        return service_model._get_metadata_property('serviceFullName')
    except UndefinedModelAttributeError:
        return None
